package sqlindexwizard.clustering;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;

import org.apache.log4j.Logger;

/**
 * Executes the KMeans Algorithm. 
 * 
 * @author Benjamin Daeumlich
 *
 */
public class KMeansAlgo {

	private static Logger logger = Logger.getLogger(KMeansAlgo.class);

	/**
	 * Initializing kMeans. Try to choose nrClusters means, first from clusterVectors-List and
	 * if not finished from qvList.
	 * 
	 * @param qvList
	 * @param clusterVectors
	 * @param nrClusters
	 * @return meanList
	 */
	public static ListOfQueryVectors init(ListOfQueryVectors qvList, ListOfQueryVectors clusterVectors, int nrClusters) {
		logger.debug("Starting initialization of kMeans!");
		
		ListOfQueryVectors meanList = new ListOfQueryVectors();
		if (clusterVectors != null) {
			meanList.addAll(getRandomQueryVectors(clusterVectors, nrClusters));
		}
		if (qvList != null) {
			nrClusters = nrClusters - meanList.size();
			meanList.addAll(getRandomQueryVectors(qvList, nrClusters));
		}
		Iterator<QueryVector> it = meanList.iterator();
		while (it.hasNext()) {
			it.next().setClusterSize(0);
		}
		logger.debug("Finished initialization of kMeans with " + meanList.size() + " means!");
		return meanList;
	}
	
	/**
	 * Choosing nrClusters QueryVectors from qvList randomly.
	 * 
	 * @param qvList
	 * @param nrClusters
	 * @return meanList
	 */
	public static ListOfQueryVectors getRandomQueryVectors(ListOfQueryVectors qvList, int nrClusters) {
		ListOfQueryVectors qvListClone = qvList.clone();
		ListOfQueryVectors meansList = new ListOfQueryVectors();
		Random r = new Random();
		for (int i = 0; i < nrClusters && 0 < qvListClone.size(); i++) {
			int randIndex = r.nextInt(qvListClone.size());
			meansList.add(qvListClone.get(randIndex));
			qvListClone.remove(randIndex);
		}
		return meansList;
	}
	
	/**
	 * Executes the kmeans algorithm.
	 * 
	 * @param QVList qvs to cluster
	 * @param initialMeans inital Cluster
	 * @return means
	 */
	public static ListOfQueryVectors start(ListOfQueryVectors QVList,
			ListOfQueryVectors initialMeans) {
		logger.debug("Starting k-Means-Algo with " + initialMeans.size() + " means and " + QVList.size() + " QueryVectors!");
		
		if (QVList == null) {
			return null;
		}
		ListOfQueryVectors meanList = initialMeans.clone();
		int meanSize = meanList.size();
		int queryListSize = QVList.size();
		// each query gets meannr
		int queryToMeanNr[] = new int[queryListSize];
		for (int i = 0; i < queryToMeanNr.length; i++) {
			queryToMeanNr[i] = -1;
		}
		// each mean gets querynrs
		ArrayList<ArrayList<Integer>> meanToQueryNr = new ArrayList<ArrayList<Integer>>(meanSize);
		for (int i = 0; i < meanSize; i++) {
			meanToQueryNr.add(new ArrayList<Integer>());
		}
		boolean finished = false;
		int nrLoops = 1;
		// staring algo
		while (!finished) {
			logger.debug("Stating Loop " + nrLoops + "!");
			int newClusters = 0;
			int queryPos = 0;
			finished = true;
			Iterator<QueryVector> it = QVList.iterator();
			while (it.hasNext()) {
				QueryVector QV = it.next();
				int clusterNr = 0;
				double sim = 0.0;
				// finding most similar cluster
				for (int i = 0; i < meanSize; i++) {
					double simComp = meanList.get(i).sim(QV);
					if (simComp > sim) {
						sim = simComp;
						clusterNr = i;
					}
				}
				// assigning qv to new cluster if necessary
				if (clusterNr != queryToMeanNr[queryPos]) {
					if (queryToMeanNr[queryPos] != -1) {
						// not first loop
						meanList.get(queryToMeanNr[queryPos]).decrementClusterSize(QV.getQ().getCount());
						int indexOfQueryPos = meanToQueryNr.get(queryToMeanNr[queryPos]).indexOf(queryPos);
						meanToQueryNr.get(queryToMeanNr[queryPos]).remove(indexOfQueryPos);
					}
					meanList.get(clusterNr).incrementClusterSize(QV.getQ().getCount());
					meanToQueryNr.get(clusterNr).add(queryPos);
					queryToMeanNr[queryPos] = clusterNr;
					newClusters++;
				}
				queryPos++;
			}
			// setting new means if necessary
			if (newClusters > 0) {
				// new Clusters found
				for (int i = 0; i < meanSize; i++) {
					ListOfQueryVectors clusterQVList = new ListOfQueryVectors();
					// generating list of QVs for each mean
					for (int j = 0; j < meanToQueryNr.get(i).size(); j++) {
						int queryId = meanToQueryNr.get(i).get(j);
						clusterQVList.add(QVList.get(queryId));
					}
					// setting new mean
					QueryVector newMean = clusterQVList.calculateMeanVector();
					if (!newMean.equals(meanList.get(i))) {
						finished = false;
						QueryVector meanOld = meanList.get(i);
						meanList.set(i, newMean);
						meanList.get(i).setClusterSize(meanOld.getClusterSize());
					}
				}
			}
			// finished one loop
			nrLoops++;
			// stopping after 1000 loops in case of infinity loop (query jumping
			// from one mean to another and vice versa
			if (nrLoops == 100) {
				finished = true;
			}
		}
		// finding real queries for means
		for (int i = 0; i < meanSize; i++) {
			ListOfQueryVectors clusterQVList = new ListOfQueryVectors();
			// generating list of QVs for each mean
			for (int j = 0; j < meanToQueryNr.get(i).size(); j++) {
				int queryId = meanToQueryNr.get(i).get(j);
				clusterQVList.add(QVList.get(queryId));
			}
			QueryVector meanQuery = clusterQVList.calculateMeanQuery();
			meanList.set(i, meanQuery);
		}
		// generating result, removing null-query-vectors
		ListOfQueryVectors resultMeanList = new ListOfQueryVectors();
		for (int i = 0; i < meanSize; i++) {
			if (!(meanList.get(i) == null)) {
				resultMeanList.add(meanList.get(i));
			}
		}
		logger.debug("Finished k-Means-Algo with " + resultMeanList.size() + " means!");
		return resultMeanList;
	}
}
