package sqlindexwizard.clustering;

import sqlindexwizard.workload.DeleteQuery;
import sqlindexwizard.workload.InsertQuery;
import sqlindexwizard.workload.ListOfQueries;
import sqlindexwizard.workload.Query;
import sqlindexwizard.workload.SelectQuery;
import sqlindexwizard.workload.UpdateQuery;
import sqlindexwizard.workload.Workload;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * A List of query Vectors, can be generated from workload or from queryList
 * 
 * @author Benjamin Daeumlich
 *
 */
public class ListOfQueryVectors extends ArrayList<QueryVector>{

	private static final long serialVersionUID = -3566572266649254596L;

	/**
	 * Constructor, calls default constructor.
	 */
	public ListOfQueryVectors() {
		super();
	}
	
	/**
	 * Generating ListOfQueryVectors from queryList.
	 * 
	 * @param queryList
	 */
	public ListOfQueryVectors(ListOfQueries queryList) {
		if (queryList == null) return;
		Iterator<Query> it = queryList.iterator();
		while (it.hasNext()) {
			Query q = it.next();
			QueryVector qv = QueryVector.newInstance(q);
			this.add(qv);
		}
	}
	
	/**
	 * Generating ListOfQueryVectors from workload.
	 * 
	 * @param w Workload
	 */
	public ListOfQueryVectors(Workload w) {
		this(w.getSelectQueryList());
		ListOfQueries qList = new ListOfQueries();
		qList.addAll(w.getSelectQueryList());
		//this(qList);
	}
	
	/**
	 * Getting list of Queries from this ListOfQueryVectors.
	 * 
	 * @return Queries
	 */
	public ListOfQueries getListOfQueries() {
		if (this == null) return null;
		ListOfQueries queryList = new ListOfQueries();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector QV = it.next();
			if (QV != null && QV.getQ() != null) {
				// set the count value of the query to the cluster size
				Query toAdd = QV.getQ().clone();
				toAdd.setCount(QV.getClusterSize());
				queryList.add(toAdd);
			}
		}
		return queryList;
	}
	
	/**
	 * Getting list of SelectQueries from this ListOfQueryVectors.
	 * 
	 * @return SelectQueries
	 */
	public ListOfQueries getListOfSelectQueries() {
		if (this == null) return null;
		ListOfQueries queryList = new ListOfQueries();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector QV = it.next();
			if (QV != null && QV.getQ() != null) {
				// set the count value of the query to the cluster size
				Query toAdd = QV.getQ().clone();
				toAdd.setCount(QV.getClusterSize());
				if (toAdd instanceof SelectQuery) {
					queryList.add(toAdd);
				}
			}
		}
		return queryList;
	}
	
	/**
	 * Getting list of UpateQueries from this ListOfQueryVectors.
	 * 
	 * @return UpdateQueries
	 */
	public ListOfQueries getListOfUpdateQueries() {
		if (this == null) return null;
		ListOfQueries queryList = new ListOfQueries();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector QV = it.next();
			if (QV != null && QV.getQ() != null) {
				// set the count value of the query to the cluster size
				Query toAdd = QV.getQ().clone();
				toAdd.setCount(QV.getClusterSize());
				if (toAdd instanceof UpdateQuery) {
					queryList.add(toAdd);
				}
			}
		}
		return queryList;
	}
	
	/**
	 * Getting list of InsertQueries from this ListOfQueryVectors.
	 * 
	 * @return InsertQueries
	 */
	public ListOfQueries getListOfInsertQueries() {
		if (this == null) return null;
		ListOfQueries queryList = new ListOfQueries();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector QV = it.next();
			if (QV != null && QV.getQ() != null) {
				// set the count value of the query to the cluster size
				Query toAdd = QV.getQ().clone();
				toAdd.setCount(QV.getClusterSize());
				if (toAdd instanceof InsertQuery) {
					queryList.add(toAdd);
				}
			}
		}
		return queryList;
	}
	
	/**
	 * Getting list of DeleteQueries from this ListOfQueryVectors.
	 * 
	 * @return DeleteQueries
	 */
	public ListOfQueries getListOfDeleteQueries() {
		if (this == null) return null;
		ListOfQueries queryList = new ListOfQueries();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector QV = it.next();
			if (QV != null && QV.getQ() != null) {
				// set the count value of the query to the cluster size
				Query toAdd = QV.getQ().clone();
				toAdd.setCount(QV.getClusterSize());
				if (toAdd instanceof DeleteQuery) {
					queryList.add(toAdd);
				}
			}
		}
		return queryList;
	}
	
	/**
	 * Calculates the mean vector from this list of query vectors.
	 * The mean vector is no real vector.
	 * 
	 * @return mean
	 */
	public QueryVector calculateMeanVector() {
		QueryVector mean = getFirstQV();
		if (mean == null || this.size() == 0) {
			return QueryVector.newInstance();
		}
		return mean.calculateMeanVector(this);
	}
	
	/**
	 * Calculates the mean query from this list of query vectors.
	 * This is the query with the highest summed up similarity.
	 * 
	 * @return mean
	 */
	public QueryVector calculateMeanQuery() {
		if (this.size() == 0) {
			return QueryVector.newInstance();
		}
		int clusterSize = 0;
		double simComp = 0.0;
		double simTotal = 0.0;
		QueryVector mean = getFirstQV();
		Iterator<QueryVector> it = this.iterator();
		while (it.hasNext()) {
			QueryVector candidate = it.next();
			simComp = 0.0;
			Iterator<QueryVector> itComp = this.iterator();
			// summing up similarities
			while (itComp.hasNext()) {
				simComp += itComp.next().sim(candidate);
			}
			// test whether better qv has been found
			if (simComp > simTotal) {
				simTotal = simComp;
				mean = candidate;
			}
			clusterSize += candidate.getQ().getCount();
		}
		mean.setClusterSize(clusterSize);
		return mean;
	}
	
	/**
	 * Getting the first QV different from null from this qv-list.
	 * 
	 * @return first valid QV
	 */
	public QueryVector getFirstQV() {
		for (int i = 0; i < this.size(); i++) {
			if (this.get(i) != null) {
				return this.get(i);
			}
		}
		return null;
	}
	
	/**
	 * Clones this list of QueryVectors.
	 * 
	 * @return clone
	 */
	public ListOfQueryVectors clone() {
		ListOfQueryVectors clone = new ListOfQueryVectors();
		for (int i = 0; i < this.size(); i++) {
			clone.add(this.get(i).clone());
		}
		return clone;
	}
	
	/**
	 * String representation of a qv list.
	 * 
	 * @return string representation
	 */
	public String toString() {
		StringBuffer result = new StringBuffer();
		result.append("QueryVectors: \n");
		Iterator<QueryVector> temp = this.iterator();
		while(temp.hasNext()) {
			result.append("\t" + temp.next().toString() + "\n");
		}
		return result.toString();
	}
	
}
