package sqlindexwizard.clustering;

import java.sql.Date;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import sqlindexwizard.dbconnection.RunstatsDB2;
import sqlindexwizard.dbconnection.SelectivityDB2;
import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.filter.JoinElement;
import sqlindexwizard.filter.SelectElement;
import sqlindexwizard.util.Config;
import sqlindexwizard.workload.Workload;

/**
 * This class gathers information needed for clustering.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class PrepareClustering {
	
	private static Logger logger = Logger.getLogger(PrepareClustering.class);

	/**
	 * Generating initial workload. Sets tables for all items in the workload, executes 
	 * runstats if needed and generates selectivities for all items from the where-clauses
	 * of the queries.
	 * 
	 * @param s files
	 * @return initial workload
	 */
	public static Workload generateInitialWorkload(String[] s) {
		logger.info("Generating initial Workload!");
		Workload w = new Workload(s);
		setItemTablesAndTypes();
		setJoinTables();
		setOtherTables();
		runRunstats();
		generateSelectivities();
		logger.info("Finished generating initial Workload!");
		return w;
	}
	
	/**
	 * Goes through the ItemList of the Workload and takes information about the types
	 * and tables of the Items from the database.
	 * 
	 */
	private static void setItemTablesAndTypes() {
		logger.info("Setting Tables and Types for where-clause-items!");
		ArrayList<ItemWhereElement> itemListOld = Workload.getItemList();
		ArrayList<ItemWhereElement> itemListNew = new ArrayList<ItemWhereElement>();
		for (int i = 0; i < itemListOld.size(); i++) {
			ItemWhereElement temp = itemListOld.get(i);
			String column = temp.getColumn().toUpperCase();
			ArrayList<String> possibleTables = temp.getPossibleTables();
			if (!temp.hasTable()) {
				for (int j = 0; j < possibleTables.size(); j++) {
					TableElement tab = Workload.getMapOfTableElements().get(possibleTables.get(j));
					if (tab == null) {
						tab = new TableElement(possibleTables.get(j));
						Workload.getMapOfTableElements().put(possibleTables.get(j), tab);
					}
					if (tab.getColumns().contains(column)) {
						temp.setTable(possibleTables.get(j));
						break;
					}
				}
			}
			if (temp.getType() == 0) {
				TableElement tab = Workload.getMapOfTableElements().get(temp.getTable());
				if (tab == null) {
					tab = new TableElement(temp.getTable());
					Workload.getMapOfTableElements().put(temp.getTable(), tab);
				}
				temp.setType(tab.getColType(column));
			}
			if (!itemListNew.contains(temp)) {
				itemListNew.add(temp);
			}
		}
		Workload.setItemList(itemListNew);
		logger.info("Finished setting Tables and Types for where-clause-items!");
	}
	
	/**
	 * Goes through the JoinList of the Workload and takes information about the 
	 * tables of the JoinElements from the database.
	 * 
	 */
	private static void setJoinTables() {
		logger.info("Setting Tables for Joins!");
		ArrayList<JoinElement> joinListOld = Workload.getJoinList();
		ArrayList<JoinElement> joinListNew = new ArrayList<JoinElement>();
		for (int i = 0; i < joinListOld.size(); i++) {
			JoinElement temp = joinListOld.get(i);
			String columnLeft = temp.getColumnLeft().toUpperCase();
			String columnRight = temp.getColumnRight().toUpperCase();
			ArrayList<String> possibleTables = temp.getPossibleTables();
			if (!temp.hasTableLeft()) {
				for (int j = 0; j < possibleTables.size(); j++) {
					TableElement tab = Workload.getMapOfTableElements().get(possibleTables.get(j));
					if (tab == null) {
						tab = new TableElement(possibleTables.get(j));
						Workload.getMapOfTableElements().put(possibleTables.get(j), tab);
					}
					if (tab.getColumns().contains(columnLeft)) {
						temp.setTableLeft(possibleTables.get(j));
						break;
					}
				}
			}
			if (!temp.hasTableRight()) {
				for (int j = 0; j < possibleTables.size(); j++) {
					TableElement tab = Workload.getMapOfTableElements().get(possibleTables.get(j));
					if (tab == null) {
						tab = new TableElement(possibleTables.get(j));
						Workload.getMapOfTableElements().put(possibleTables.get(j), tab);
					}
					if (tab.getColumns().contains(columnRight)) {
						temp.setTableRight(possibleTables.get(j));
						break;
					}
				}
			}
			if (!joinListNew.contains(temp)) {
				joinListNew.add(temp);
			}
		}
		Workload.setJoinList(joinListNew);
		logger.info("Finished setting Tables for Joins!");
	}
	
	/**
	 * Goes through the OtherList of the Workload and takes information about the 
	 * tables of the other SelectElements from the database.
	 * 
	 */
	private static void setOtherTables() {
		logger.info("Setting Tables for other items (select, order by, group by)!");
		ArrayList<SelectElement> otherListOld = Workload.getOtherList();
		ArrayList<SelectElement> otherListNew = new ArrayList<SelectElement>();
		for (int i = 0; i < otherListOld.size(); i++) {
			SelectElement temp = otherListOld.get(i);
			if (temp.isWildcard()) {
				continue;
			}
			String column = temp.getColumn().toUpperCase();
			ArrayList<String> possibleTables = temp.getPossibleTables();
			if (!temp.hasTable()) {
				for (int j = 0; j < possibleTables.size(); j++) {
					TableElement tab = Workload.getMapOfTableElements().get(possibleTables.get(j));
					if (tab == null) {
						tab = new TableElement(possibleTables.get(j));
						Workload.getMapOfTableElements().put(possibleTables.get(j), tab);
					}
					if (tab.getColumns().contains(column)) {
						temp.setTable(possibleTables.get(j));
						break;
					}
				}
			}
			if (!otherListNew.contains(temp)) {
				otherListNew.add(temp);
			}
		}
		Workload.setOtherList(otherListNew);
		logger.info("Finished setting Tables for other items (select, order by, group by)!");
	}
	
	/**
	 * Executes RUNSTATS if it is necessary.
	 * 
	 */
	private static void runRunstats() {
		logger.info("Checking if execution of runstats is needed!");
		ArrayList<String> tables = new ArrayList<String>(Workload.getMapOfTableElements().keySet());
		RunstatsDB2 runstats = new RunstatsDB2();
		if (runstats.getExecute()) {
			for (int i = 0; i < tables.size(); i++) {
				runstats.addRunstatsCmd(tables.get(i));
			}
			runstats.run();
			Config.setPropertyString("db2.lastRunstats", new Date(System.currentTimeMillis()).toString());
		} else {
			logger.info("No execution of runstats needed!");
		}
		logger.info("Finished checking if execution of runstats is needed!");
	}
	
	/**
	 * Generates selectivities for all Elements of the ItemList of the Workload.
	 * 
	 */
	private static void generateSelectivities() {
		logger.info("Generating Selectivities!");
		SelectivityDB2 sel = new SelectivityDB2();
		ArrayList<ItemWhereElement> itemList = Workload.getItemList();
		for (int i = 0; i < itemList.size(); i++) {
			ItemWhereElement temp = itemList.get(i);
			TableElement tab = Workload.getMapOfTableElements().get(temp.getTable());
			if (tab == null) {
				tab = new TableElement(temp.getTable());
				Workload.getMapOfTableElements().put(temp.getTable(),tab);
			}
			if (Workload.getMapOfSelectivities().get(temp) == null) {
				double selectivity = sel.getSelectivity(temp, tab);
				if (temp.isNot()) {
					Workload.getMapOfSelectivities().put(temp, 1-selectivity);
				} else {
					Workload.getMapOfSelectivities().put(temp, selectivity);
				}
			}
			
		}
		sel.finalize();
		logger.info("Finished generating Selectivities!");
	}
	
	/**
	 * Printing debug information.
	 * 
	 */
	public static void printDebug() {
		System.out.println("Debug information: \n");
		System.out.println("ItemList:\t" + Workload.getItemList());
		System.out.println("JoinList:\t" + Workload.getJoinList());
		System.out.println("OtherList:\t" + Workload.getOtherList());
		System.out.println("SelList:\t" + Workload.getMapOfSelectivities());
		System.out.println("TableElements:\n\n" + Workload.getMapOfTableElements());
	}
}
