package sqlindexwizard.clustering;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import sqlindexwizard.application.Input;
import sqlindexwizard.workload.Query;

/**
 * Representation of a query vector.
 * 
 * @author Benjamin Daeumlich
 *
 */
public abstract class QueryVector {
	
	protected Query q;
	protected int clusterSize;
	protected double length;
	protected HashMap<QueryVectorItem, Double> QVMap = new HashMap<QueryVectorItem, Double>();
	
	protected final boolean countSameItemsNumerous = Input.countSameItemsNumerous;
	
	protected static Logger logger = Logger.getLogger(QueryVector.class);
	
	/**
	 * Constructor generates empty query vector.
	 * 
	 */
	public QueryVector() {
		logger.debug("Generation empty query vector.");
		this.q = null;
		this.clusterSize = 0;
		this.length = 0;
	}
	
	/**
	 * Constructor generates query vector from Query.
	 * 
	 * @param q
	 */
	public QueryVector(Query q) {
		logger.debug("Generating query vector from query " + q.toString() + "!");
		this.q = q;
		this.clusterSize = q.getCount();
		this.length = 0;
		generateQueryVector(q);
	}

	/**
	 * Constructor generates query vector from a query vector (a clone).
	 * 
	 * @param qv
	 */
	public QueryVector(QueryVector qv) {
		this.q = qv.getQ().clone();
		this.clusterSize = qv.clusterSize;
		this.length = qv.length;
		for (Entry<QueryVectorItem, Double> entry : qv.QVMap.entrySet()) {
			QVMap.put(entry.getKey(), entry.getValue());
		}
	}
	
	/**
	 * Generates query vector from Query.
	 * 
	 * @param q
	 */
	public abstract void generateQueryVector(Query q);
	
	/**
	 * Calculates the length of the query-vector.
	 * 
	 * @return length
	 */
	public abstract double calculateLength();
	
	/**
	 * Adding value to qvmap. calculates new length.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public abstract void addValue(QueryVectorItem it, double sel);
	
	/**
	 * Adding value to qvmap for summation of qv. calculates new length.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public abstract void addValueForSum(QueryVectorItem it, double sel);
	
	/**
	 * Generates VectorEntry 1-selectivity.
	 * 
	 * @param selectivity
	 * @return 1-selectivity
	 */
	protected double generateVectorEntryFromSelectivity(double selectivity) {
		return 1 - selectivity;
	}
	
	/**
	 * Tests if vector is nullVector.
	 * 
	 * @return true if nullvector
	 */
	public boolean isNullVector() {
		return (this.length == 0);
	}
	
	/**
	 * Computes similarity between this qv and another qv (angle(this,qv)).
	 * 
	 * @param qv other query vector
	 * @return similarity
	 */
	public abstract double sim(QueryVector qv);
	
	/**
	 * Computing scalar product between this queryVector and another queryvector.
	 * 
	 * @param qv other qv
	 * @return scalar product
	 */
	public abstract double computeScalarProduct(QueryVector qv);
	
	/**
	 * Calculates the mean vector from a list of query vectors.
	 * The mean vector is no real vector.
	 * 
	 * @param qvList
	 * @return mean
	 */
	public abstract QueryVector calculateMeanVector(ListOfQueryVectors qvList);
	
	/**
	 * Sums up all Query Vectors from qvList.
	 * 
	 * @param qvList
	 * @return sumVector
	 */
	public abstract QueryVector sumUpQueryVectors(ListOfQueryVectors qvList);
	
	/**
	 * Incrementing clustersize.
	 * 
	 * @param value value to increment
	 */
	public void incrementClusterSize(int value) {
		clusterSize += value;
	}
	
	/**
	 * Decrementing clustersize.
	 * 
	 * @param value value to decrement
	 */
	public void decrementClusterSize(int value) {
		clusterSize -= value;
	}
	
	/**
	 * Getting the Query which the query vector represents.
	 * 
	 * @return query
	 */
	public Query getQ() {
		return q;
	}
	
	/**
	 * Getting the clusterSize
	 * 
	 * @return the clusterSize
	 */
	public int getClusterSize() {
		return clusterSize;
	}
	
	/**
	 * Setting the clusterSize.
	 * 
	 * @param clusterSize the clusterSize to set
	 */
	public void setClusterSize(int clusterSize) {
		this.clusterSize = clusterSize;
	}

	/**
	 * Generates a query vector.
	 * 
	 * @return a new query vector object
	 */
	public static QueryVector newInstance() {
		try {
			Class<?> clazz = Input.QUERYVECTOR_CLASS;
			return (QueryVector) (clazz.getConstructor().newInstance());
		} catch (Exception e) {
			logger.error("Cannot generate new Instance of QueryVector with class " + Input.QUERYVECTOR_CLASS + "! Error: " + e.toString());
			return null;
		}
	}
	
	/**
	 * Generates a query vector.
	 * 
	 * @param q a Query
	 * @return a new query vector object
	 */
	public static QueryVector newInstance(Query q) {
		try {
			Class<?> clazz = Input.QUERYVECTOR_CLASS;
			return (QueryVector) (clazz.getConstructor(Query.class).newInstance(q));
		} catch (Exception e) {
			logger.error("Cannot generate new Instance of QueryVector with class " + Input.QUERYVECTOR_CLASS + "! Error: " + e.toString());
			return null;
		}
	}
	
	/**
	 * Clones the QueryVector.
	 * 
	 * @return clone 
	 */
	public abstract QueryVector clone();
	
	/**
	 * Tests for equality of two Objects.
	 * 
	 * @return true if they are equal
	 */
	public boolean equals(Object o) {
		if (o instanceof QueryVector) {
			return equals((QueryVector) o);
		} else {
			return false;
		}
	}

	/**
	 * Tests for equality of two Objects.
	 * 
	 * @return true if they are equal
	 */
	public abstract boolean equals(QueryVector qv);

	/**
	 * Tests for equality of two qv-maps.
	 * 
	 * @param qv1
	 * @param qv2
	 * @return true if qv1 and qv2 are equal
	 */
	public boolean equalsQVMap(HashMap<QueryVectorItem,Double> qv1, HashMap<QueryVectorItem, Double> qv2) {
		if (qv1.size() != qv2.size()) {
			return false;
		}
		Iterator<Entry<QueryVectorItem,Double>> it = qv1.entrySet().iterator();
		while (it.hasNext()) {
			Entry<QueryVectorItem,Double> entry = it.next();
			QueryVectorItem key = entry.getKey();
			Double value = entry.getValue();
			if (value == null) {
				if (!(qv2.get(key) == null && qv2.containsKey(key))) {
					return false;
				}
			} else {
				Double valueComp = qv2.get(key);
				if (valueComp == null) {
					return false;
				}
				if (value != valueComp) {
					return false;
				}
			}
		}
		return true;
	}
	
	/**
	 * String representation of the query vector.
	 * 
	 * @return string representation
	 */
	public String toString() {
		return QVMap.toString();
	}

}