package sqlindexwizard.clustering;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.workload.Query;
import sqlindexwizard.workload.Workload;

/**
 * Representation of a query Vector.
 * 
 * Properties: 	- 2 dimensions per column (1 and sel)
 * 				- for same colums: (1-(sel(a=1)+sel(a=2)))
 * 
 * @author Benjamin Daeumlich
 *
 */
public class QueryVectorSim3 extends QueryVector {
	
	private HashMap<QueryVectorItem, Double> QVMap2;

	/**
	 * Constructor, calls super constructor.
	 * 
	 */
	public QueryVectorSim3() {
		super();
		QVMap2 = new HashMap<QueryVectorItem, Double>();
	}

	/**
	 * Constructor, calls super constructor with query as parameter.
	 * 
	 * @param q query
	 */
	public QueryVectorSim3(Query q) {
		super(q);
	}
	
	/**
	 * Constructor generates query vector from a query vector (a clone).
	 * 
	 * @param qv
	 */
	public QueryVectorSim3(QueryVectorSim3 qv) {
		super(qv);
		QVMap2 = new HashMap<QueryVectorItem, Double>();
		for (Entry<QueryVectorItem, Double> entry : qv.QVMap2.entrySet()) {
			QVMap2.put(entry.getKey(), entry.getValue());
		}
	}
	
	/**
	 * Generates query vector from Query.
	 * 
	 * @param q
	 */
	public void generateQueryVector(Query q) {
		QVMap2 = new HashMap<QueryVectorItem, Double>();
		ArrayList<String> tables = q.getTables();
		if (tables == null) {
			logger.error("Trying to generate query vector from invalid query.");
			return;
		}
		
		for (int i = 0; i < tables.size(); i++) {
			TableElement tab = Workload.getMapOfTableElements().get(tables.get(i));
			if (tab == null) {
				tab = new TableElement(tables.get(i));
				Workload.getMapOfTableElements().put(tables.get(i), tab);
			}
			ArrayList<String> columns = tab.getColumns();
			for (int j=0; j < columns.size(); j++) {
				QueryVectorItem temp = new QueryVectorItem(tables.get(i),columns.get(j));
				QVMap.put(temp, 0.0);
				QVMap2.put(temp, 0.0);
			}
		}
		List<ItemWhereElement> itemList = q.getItemList();
		if (itemList == null) {
			logger.error("Trying to generate query vector from invalid query with no items.");
			return;
		}
		if (!countSameItemsNumerous) {
			List<ItemWhereElement> itemListTemp = new ArrayList<ItemWhereElement>();
			for (int i = 0; i < itemList.size(); i++) {
				if (!itemListTemp.contains((ItemWhereElement)itemList.get(i))) {
					itemListTemp.add((ItemWhereElement)itemList.get(i));
				}
			}
			itemList = itemListTemp;
		}
		Iterator<ItemWhereElement> it = itemList.iterator();
		while (it.hasNext()) {
			ItemWhereElement temp = it.next();
			double sel = Workload.getMapOfSelectivities().get(temp);
			
			addValue(new QueryVectorItem(temp.getTable(),temp.getColumn()), sel*temp.getCount());
		}
		this.length = Math.sqrt(length);
	}
	
	/**
	 * Adding value to qvmap. calculates new length.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public void addValue(QueryVectorItem it, double sel) {
		double selOld = QVMap.get(it);
		double selOld2 = QVMap2.get(it);
		double selNew = 1 - selOld + sel;
		if (selNew > 1) { 
			selNew -= 1;
		}
		selNew = generateVectorEntryFromSelectivity(selNew);
		QVMap.put(it, selNew);
		QVMap2.put(it, 1.0);
		this.length = this.length - selOld*selOld - selOld2*selOld2 + selNew*selNew + 1;
	}
	
	/**
	 * Computing scalar product between this queryVector and another queryvector.
	 * 
	 * @param qv other qv
	 * @return scalar product
	 */
	public double computeScalarProduct(QueryVector qv) {
		double result = 0;
		Iterator<QueryVectorItem> items = QVMap.keySet().iterator();
		while (items.hasNext()) {
			QueryVectorItem temp = items.next();			
			result += (this.QVMap.get(temp)*qv.QVMap.get(temp) + this.QVMap2.get(temp)*((QueryVectorSim3)qv).QVMap2.get(temp));
		}
		return result;
	}
	
	/**
	 * Computes similarity between this qv and another qv (angle(this,qv)).
	 * 
	 * @param qv other query vector
	 * @return similarity
	 */
	public double sim(QueryVector qv) {
		if (this.isNullVector() && qv.isNullVector()) {
			return 1.0;
		}
		if (this.isNullVector()) {
			return 0.0;
		}
		if (qv.isNullVector()) {
			return 0.0;
		}
		double numerator = this.computeScalarProduct(qv);
		double denominator = this.calculateLength() * qv.calculateLength();
		if (denominator != 0) {
			return numerator / denominator;
		}
		return 0.0;
	}

	/**
	 * Getting QVMap2.
	 * 
	 * @return the qVMap2
	 */
	public HashMap<QueryVectorItem, Double> getQVMap2() {
		return QVMap2;
	}
	
	/**
	 * Clones the QueryVector.
	 * 
	 * @return clone 
	 */
	public QueryVectorSim3 clone() {
		return new QueryVectorSim3(this);
	}
	
	/**
	 * Calculates the mean vector from a list of query vectors.
	 * The mean vector is no real vector.
	 * 
	 * @param qvList
	 * @return mean
	 */
	public QueryVector calculateMeanVector(ListOfQueryVectors qvList) {
		int size = qvList.size();
		if (size == 0) {
			return new QueryVectorSim3();
		}
		
		QueryVectorSim3 mean = sumUpQueryVectors(qvList);
		for (QueryVectorItem item: mean.QVMap.keySet()) {
			double value = mean.QVMap.get(item);
			value = value / size;
			mean.QVMap.put(item, value);
		}
		mean.length = mean.length / size;
		return mean;
	}

	/**
	 * Tests for equality of two Objects.
	 * 
	 * @return true if they are equal
	 */
	public boolean equals(QueryVector qv) {
		if (this == qv) {
			return true;
		}
		if (qv == null) {
			return false;
		}
		if (getClass() != qv.getClass()) {
			return false;
		}
		QueryVectorSim3 qvTemp = (QueryVectorSim3) qv;
		if (!equalsQVMap(this.QVMap, qvTemp.QVMap)) {
			return false;
		}
		if (!equalsQVMap(this.getQVMap2(), qvTemp.getQVMap2())) {
			return false;
		}
		return true;
	}

	/**
	 * Sums up all Query Vectors from qvList.
	 * 
	 * @param qvList
	 * @return sumVector
	 */
	public QueryVectorSim3 sumUpQueryVectors(ListOfQueryVectors qvList) {
		if (qvList == null) {
			return null;
		}
		QueryVectorSim3 sumVector = this.clone();
		for (Entry<QueryVectorItem, Double> entry: sumVector.QVMap.entrySet()) {
			sumVector.QVMap.put(entry.getKey(), 0.0);
		}
		Iterator<QueryVector> it = qvList.iterator();
		while (it.hasNext()) {
			QueryVector qv = it.next();
			if (!(qv instanceof QueryVectorSim3)) {
				logger.error("Error: Comparing different instances of Query Vectors!");
			}
			for (Entry<QueryVectorItem, Double> entry: qv.QVMap.entrySet()) {
				if (entry.getValue() != 0.0) {
					sumVector.addValueForSum(entry.getKey(),entry.getValue());
				}
			}
		}
		sumVector.length = Math.sqrt(sumVector.length);
		return sumVector;
	}

	/**
	 * Adding value to qvmap for summation of qv. calculates new length.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public void addValueForSum(QueryVectorItem it, double sel) {
		double selOld = QVMap.get(it);
		double selOld2 = QVMap2.get(it);
		double selNew = selOld + sel;
		QVMap.put(it, selNew);
		QVMap2.put(it, 1.0);
		this.length = this.length - selOld*selOld - selOld2*selOld2 + selNew*selNew + 1;
		
	}
	
	/**
	 * Calculates the length of the query-vector.
	 * 
	 * @return length
	 */
	public double calculateLength() {
		double result = 0.0;
		for (Entry<QueryVectorItem, Double> entry: this.QVMap.entrySet()) {
			if (entry.getValue() != 0.0) {
				result += (entry.getValue()*entry.getValue());
			}
		}
		for (Entry<QueryVectorItem, Double> entry: this.QVMap2.entrySet()) {
			if (entry.getValue() != 0.0) {
				result += (entry.getValue()*entry.getValue());
			}
		}
		return Math.sqrt(result);
	}

}
