package sqlindexwizard.clustering;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.workload.Query;
import sqlindexwizard.workload.SelectQuery;
import sqlindexwizard.workload.Workload;

/**
 * Representation of a query Vector.
 * 
 * Using sim-method from Kolaczkowski.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class QueryVectorSim4 extends QueryVector {

	protected HashMap<ItemWhereElement, Double> QVMap2;

	/**
	 * Constructor, calls super constructor.
	 * 
	 */
	public QueryVectorSim4() {
		super();
		this.length = 1;
	}

	/**
	 * Constructor, calls super constructor with query as parameter.
	 * 
	 * @param q query
	 */
	public QueryVectorSim4(Query q) {
		super(q);
		this.length = 1;
	}
	
	/**
	 * Constructor generates query vector from a query vector (a clone).
	 * 
	 * @param qv
	 */
	public QueryVectorSim4(QueryVectorSim4 qv) {
		super(qv);
		QVMap2 = new HashMap<ItemWhereElement, Double>();
		for (Entry<ItemWhereElement, Double> entry : qv.QVMap2.entrySet()) {
			QVMap2.put(entry.getKey(), entry.getValue());
		}
		this.length = 1;
	}
	
	/**
	 * Generates query vector from Query.
	 * 
	 * @param q
	 */
	public void generateQueryVector(Query q) {
		QVMap2 = new HashMap<ItemWhereElement, Double>();
		ArrayList<String> tables = q.getTables();
		if (tables == null) {
			logger.error("Trying to generate query vector from invalid query.");
			return;
		}
		for (int i = 0; i < tables.size(); i++) {
			TableElement tab = Workload.getMapOfTableElements().get(tables.get(i));
			if (tab == null) {
				tab = new TableElement(tables.get(i));
				Workload.getMapOfTableElements().put(tables.get(i), tab);
			}
			ArrayList<String> columns = tab.getColumns();
			for (int j=0; j < columns.size(); j++) {
				QVMap.put(new QueryVectorItem(tables.get(i),columns.get(j)), 0.0);
			}
		}
		List<ItemWhereElement> itemList = q.getItemList();
		if (itemList == null) {
			logger.error("Trying to generate query vector from invalid query with no items.");
			return;
		}
//		if (!countSameItemsNumerous) {
//			List<ItemWhereElement> itemListTemp = new ArrayList<ItemWhereElement>();
//			for (int i = 0; i < itemList.size(); i++) {
//				if (!itemListTemp.contains((ItemWhereElement)itemList.get(i))) {
//					itemListTemp.add((ItemWhereElement)itemList.get(i));
//				}
//			}
//			itemList = itemListTemp;
//		}
		Iterator<ItemWhereElement> it = itemList.iterator();
		while (it.hasNext()) {
			ItemWhereElement temp = it.next();
			double sel = generateVectorEntryFromSelectivity(Workload.getMapOfSelectivities().get(temp));
			QVMap2.put(temp, sel);
		}
	}
	
	/**
	 * Adding value to qvmap. calculates new length.
	 * Nothing to do for this kind of sim-calculation.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public void addValue(QueryVectorItem it, double sel) {
		// nothing to do because not needed for this kind of sim-calculation
	}
	
	/**
	 * Computing scalar product between this queryVector and another queryvector.
	 * Not needed for this kind of sim-calculation.
	 * 
	 * @param qv other qv
	 * @return scalar product
	 */
	public double computeScalarProduct(QueryVector qv) {
		return 0.0;
	}
	
	/**
	 * Computes similarity between this qv and another qv.
	 * Using sim-method of Kolaczkowski.
	 * 
	 * @param qv other query vector
	 * @return similarity
	 */
	public double sim(QueryVector qv) {
		if (!((SelectQuery)this.getQ()).getWhere().equalsSim(((SelectQuery)qv.getQ()).getWhere())) {
			return 0.0;
		}
		ArrayList<ItemWhereElement> qvItems = new ArrayList<ItemWhereElement>();
		qvItems.addAll(((QueryVectorSim4)qv).getQVMap2().keySet());
		ArrayList<ItemWhereElement> thisItems = new ArrayList<ItemWhereElement>();
		thisItems.addAll(this.QVMap2.keySet());
		double maxValue = 0.0;
		for (int i = 0; i < qvItems.size(); i++) {
			ItemWhereElement itemi = qvItems.get(i);
			double maxValueNew = 0.0;
			double seli = ((QueryVectorSim4)qv).getQVMap2().get(itemi);
			for (int j = 0; j < thisItems.size(); j++) {
				ItemWhereElement itemj = thisItems.get(j);
				if (itemi.equalsSim(itemj)) {
					double selj = this.QVMap2.get(itemj);
					if (seli == 0.0 || selj == 0.0) {
						return 0.0;
					}
					if (seli >= selj) {
						maxValueNew = seli/selj - 1.0;
					} else {
						maxValueNew = selj/seli - 1.0;
					}
				}
			}
			if (maxValueNew > maxValue) {
				maxValue = maxValueNew;
			}
		}
		return 1.0/(maxValue+1.0);
	}
	
	/**
	 * Getting QVMap2.
	 * 
	 * @return the qVMap2
	 */
	public HashMap<ItemWhereElement, Double> getQVMap2() {
		return QVMap2;
	}
	
	/**
	 * Clones the QueryVector.
	 * 
	 * @return clone 
	 */
	public QueryVectorSim4 clone() {
		return new QueryVectorSim4(this);
	}

	/**
	 * Dummy-method because not needed for this kind of sim-calculation.
	 * 
	 * @param it item to add
	 * @param sel selectivity
	 */
	public void addValueForSum(QueryVectorItem it, double sel) {
		// nothing to do because not needed for this kind of sim-calculation
	}

	/**
	 * Dummy-method because not needed for this kind of sim-calculation.
	 * 
	 * @param qvList
	 * @return this
	 */
	public QueryVector calculateMeanVector(ListOfQueryVectors qvList) {
		// nothing to do because not needed for this kind of sim-calculation
		return this;
	}

	/**
	 * Dummy-method because not needed for this kind of sim-calculation.
	 * 
	 * @param qv
	 * @return true
	 */
	public boolean equals(QueryVector qv) {
		// nothing to do because not needed for this kind of sim-calculation
		return true;
	}

	/**
	 * Dummy-method because not needed for this kind of sim-calculation.
	 * Returning dummy-value.
	 * 
	 * @param qvList
	 * @return null
	 */
	public QueryVector sumUpQueryVectors(ListOfQueryVectors qvList) {
		// TODO Auto-generated method stub
		return null;
	}
	
	/**
	 * Dummy-method because not needed for this kind of sim-calculation.
	 * 
	 * @return 1.0
	 */
	public double calculateLength() {
		return 1.0;
	}

}
