package sqlindexwizard.clustering;

import org.apache.log4j.Logger;

import sqlindexwizard.application.Input;
import sqlindexwizard.workload.Workload;

/**
 * This class compresses a workload.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class WorkloadCompression {

	private Workload WInitial;
	private Workload WCompressed;
	private ListOfQueryVectors clusterVectors = new ListOfQueryVectors();
	
	private static Logger logger = Logger.getLogger(WorkloadCompression.class);
	
	/**
	 * Constructor generates initial workload from files.
	 * 
	 * @param files
	 */
	public WorkloadCompression(String[] files) {
		WInitial = PrepareClustering.generateInitialWorkload(files);
		//PrepareClustering.printDebug();
	}
	
	/**
	 * Performing clustering. First of all do partitioning and then execute
	 * Clustering-Algorithm for each partition.
	 * 
	 */
	public void goClustering() {
		logger.info("Starting clustering with " + WInitial.getTotalNrOfQueries() + " Queries!");
		ListOfPartitions partitions = new ListOfPartitions(new ListOfQueryVectors(WInitial));
		for (int i = 0; i < partitions.size(); i++) {
			if (Input.leader && !Input.kmeans) {
				clusterVectors.addAll(LeaderAlgo.start(partitions.get(i).getQvList(), null, Input.boundValue));
			} else if (Input.kmeans && Input.leader) {
				ListOfQueryVectors leaderList = LeaderAlgo.start(partitions.get(i).getQvList(), null, Input.boundValue);
				int k = leaderList.size();
				ListOfQueryVectors kMeansInitialization = KMeansAlgo.init(partitions.get(i).getQvList(), leaderList, k);
				clusterVectors.addAll(KMeansAlgo.start(partitions.get(i).getQvList(), kMeansInitialization));
			} else if (Input.kmeans && !Input.leader) {
				int k = Math.round(partitions.get(i).getQvList().size()*Input.kValue/WInitial.getNrOfAddedQueries()) + 1;
				ListOfQueryVectors kMeansInitialization = KMeansAlgo.init(partitions.get(i).getQvList(), null, k);
				clusterVectors.addAll(KMeansAlgo.start(partitions.get(i).getQvList(), kMeansInitialization));
			} else {
				logger.error("Unknown clustering Algorithm");
			}
		}
		WCompressed = WInitial.clone();
		WCompressed.setSelectQueryList(clusterVectors.getListOfSelectQueries());
		WCompressed.setUpdateQueryList(clusterVectors.getListOfUpdateQueries());
		WCompressed.setInsertQueryList(clusterVectors.getListOfInsertQueries());
		WCompressed.setDeleteQueryList(clusterVectors.getListOfDeleteQueries());
		logger.info("Finished clustering with " + WCompressed.getTotalNrOfQueries() + " Queries!");
	}

	/**
	 * Getting initial Workload.
	 * 
	 * @return the wInitial
	 */
	public Workload getWInitial() {
		return WInitial;
	}

	/**
	 * Getting compressed Workload.
	 * 
	 * @return the wCompressed
	 */
	public Workload getWCompressed() {
		return WCompressed;
	}

	/**
	 * Getting clustervectors.
	 * 
	 * @return the clusterVectors
	 */
	public ListOfQueryVectors getClusterVectors() {
		return clusterVectors;
	}
	
	/**
	 * Getting clustering ratio #QueriesWCompressed/#QueriesWInitial.
	 * 
	 * @return clustering ratio
	 */
	public double getClusteringRation() {
		return 1.0*WCompressed.getTotalNrOfQueries()/WInitial.getTotalNrOfQueries();
	}
	
	/**
	 * String representation of Workload Compression.
	 * 
	 * @return String representation
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		if (Input.leader && !Input.kmeans) {
			sb.append("Algorithm:\tLeader\n");
			sb.append("BoundValue:\t" + Input.boundValue + "\n\n");
		} else if (Input.kmeans && Input.leader) {
			sb.append("Algorithm:\tk-Means with Leader\n");
			sb.append("BoundValue:\t" + Input.boundValue + "\n");
			sb.append("k:\t\t" + getWCompressed().getTotalNrOfQueries() + "\n\n");
		} else if (Input.kmeans && !Input.leader) {
			sb.append("Algorithm:\tk-Means\n");
			sb.append("k:\t\t" + Input.kValue + "\n\n");
		}
		sb.append("initial Workload:\n\n");
		sb.append(getWInitial() + "\n");
		sb.append("compressed Workload:\n\n");
		sb.append(getWCompressed() + "\n");
		sb.append("Clustering Ratio: " + getWCompressed().getTotalNrOfQueries() + "/" + getWInitial().getTotalNrOfQueries() + " = " + getClusteringRation() + "\n\n");
		return sb.toString();
	}

}
