package sqlindexwizard.dbconnection;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import sqlindexwizard.indexadvisor.Index;
import sqlindexwizard.indexadvisor.ListOfIndexes;

/**
 * This class provides needed functions used to get information from the database for
 * the index advisor.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class IndexDB2 extends ConnectionDB2 {
	
	private static Logger logger = Logger.getLogger(IndexDB2.class);

	/**
	 * Deletes all entries from ADVISE_INDES.
	 * 
	 */
	public void cleanAdviseIndex() {
		logger.debug("Cleaning ADVISE_INDEX table!");
		String query = "DELETE FROM ADVISE_INDEX";
		updateQuery(query);
	}
	
	/**
	 * Creates an entry in the ADVISE_INDEX table for a colum of a table.
	 * 
	 * @param column column
	 * @param table table
	 */
	public void createAdviseIndexEntry(String column, String table) {
		logger.debug("Creating Entry in ADVISE_INDEX for column " + column + " in table " + table + "!");
		String query = "SELECT " + column + " FROM " + table;
		executeQueryNoResult(query);
	}
	
	/**
	 * Sets EXPLAIN MODE to RECOMMEND INDEXES (index candidates).
	 * 
	 */
	public void setExplainModeRecommend() {
		logger.debug("Setting EXPLAIN MODE to RECOMMEND INDEXES!");
		String query = "SET CURRENT EXPLAIN MODE RECOMMEND INDEXES";
		updateQuery(query);
	}
	
	/**
	 * Sets EXPLAIN MODE to EXPLAIN (cost for query without virtual indexes).
	 * 
	 */
	public void setExplainModeExplain() {
		logger.debug("Setting EXPLAIN MODE to EXPLAIN!");
		String query = "SET CURRENT EXPLAIN MODE EXPLAIN";
		updateQuery(query);
	}
	
	/**
	 * Sets EXPLAIN MODE to EVALUATE INDEXES (cost for query with virtual indexes
	 * and costs).
	 * 
	 */
	public void setExplainModeEvaluate() {
		logger.debug("Setting EXPLAIN MODE to EVALUATE INDEXES!");
		String query = "SET CURRENT EXPLAIN MODE EVALUATE INDEXES";
		updateQuery(query);
	}
	
	/**
	 * Sets EXPLAIN MODE to NO (normal mode).
	 * 
	 */
	public void setExplainModeNo() {
		logger.debug("Setting EXPLAIN MODE to NO!");
		String query = "SET CURRENT EXPLAIN MODE NO";
		updateQuery(query);
	}
	
	/**
	 * Getting the cost of a given query q (with or without virtual indexes, depends on
	 * the recommend mode.
	 * 
	 * @param q Querystring
	 * @return costs
	 */
	public double getQueryCost(String q) {
		logger.debug("Getting cost for query " + q + "!");
		double result = 0.0;
		int errorCode = executeQueryNoResult(q);
		if (errorCode == 1) {
			return -9999999999.0;
		}
		setExplainModeNo();
		try {
			String query = "SELECT MAX(TOTAL_COST) AS COST from EXPLAIN_STATEMENT where EXPLAIN_TIME = (SELECT MAX(EXPLAIN_TIME) FROM EXPLAIN_STATEMENT)";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				result = rs.getDouble("COST");
			}
			rs.close();
			closeStatement();
		} catch (SQLException e) {
			logger.error("Error in getQueryCost(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getQueryCost(): " + e.toString());
		}
		return result;
	}
	
	/**
	 * Return the Index with needed properties for indexName.
	 * 
	 * @param indexName
	 * @return index with information
	 */
	public Index getIndex(String indexName) {
		logger.debug("Getting Index with Name " + indexName + "!");
		Index idx = new Index(indexName);
		try {
			String query = "SELECT TBNAME, COLNAMES, CREATION_TEXT, NLEAF, USERDEFINED, SYSTEM_REQUIRED, EXISTS FROM ADVISE_INDEX WHERE NAME = '" + indexName + "'";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				idx.setIndexCreationText(rs.getString("CREATION_TEXT"));
				String table = rs.getString("TBNAME");
				String columns = rs.getString("COLNAMES");
				columns = columns.substring(1, columns.length());
				idx.setIndexTableAndColumn(table + "." + columns);
				//idx.setNrLeafInitial(rs.getLong("NLEAF"));
				int userDefined = rs.getInt("USERDEFINED");
				int systemRequired = rs.getInt("SYSTEM_REQUIRED");
				if (userDefined == 1) {
					idx.setUserDefined(true);
				}
				if (systemRequired == 1) {
					idx.setSystemRequired(true);
				}
				char exists = rs.getString("EXISTS").charAt(0);
				if (exists == 'Y') {
					idx.setExists(true);
				}
				//double size = Double.valueOf(rs.getString("NLEAF")) * 4096.0 / (1024.0*1024.0);
				//idx.setSize(size);
			}
			rs.close();
			closeStatement();			
		} catch (SQLException e) {
			logger.error("Error in getIndex(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getIndex(): " + e.toString());
		}
		return idx;
	}
	
	/**
	 * Generates the List of indexes with all possible indexes from ADVISE_INDEX with
	 * their properties.
	 * 
	 * @return IndexList with properties
	 */
	public ListOfIndexes getPossibleIndexes() {
		logger.debug("Getting all possible Indexes from ADVISE_INDEX!");
		ListOfIndexes result = new ListOfIndexes();
		try {
			String query = "SELECT NAME, TBNAME, COLNAMES, CREATION_TEXT, NLEAF, USERDEFINED, SYSTEM_REQUIRED, EXISTS FROM ADVISE_INDEX";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				String indexName = rs.getString("NAME");
				Index idx = new Index(indexName);
				idx.setIndexCreationText(rs.getString("CREATION_TEXT"));
				String table = rs.getString("TBNAME");
				String columns = rs.getString("COLNAMES");
				columns = columns.substring(1, columns.length());
				idx.setIndexTableAndColumn(table + "." + columns);
				idx.setNrLeafInitial(rs.getLong("NLEAF"));
				int userDefined = rs.getInt("USERDEFINED");
				int systemRequired = rs.getInt("SYSTEM_REQUIRED");
				if (userDefined == 1) {
					idx.setUserDefined(true);
				}
				if (systemRequired == 1) {
					idx.setSystemRequired(true);
				}
				char exists = rs.getString("EXISTS").charAt(0);
				if (exists == 'Y') {
					idx.setExists(true);
				}
				double size = Double.valueOf(rs.getString("NLEAF")) * 4096.0 / (1024.0*1024.0);
				if (size > 0) {
					idx.setSize(size);
				}
				if (!result.contains(idx)) {
					result.add(idx);
				}
			}
			rs.close();
			closeStatement();			
		} catch (SQLException e) {
			logger.error("Error in getPossibleIndexes(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getPossibleIndexes(): " + e.toString());
		}
		return result;
	}
	
	/**
	 * Getting the used indexes for the last executed query.
	 * 
	 * @return used indexes
	 */
	public ListOfIndexes getUsedIndexes() {
		logger.debug("Getting used Indexes for last query!");
		ListOfIndexes result = new ListOfIndexes();
		try {
			String query = "SELECT OBJECT_NAME, NLEAF from EXPLAIN_OBJECT where EXPLAIN_TIME = (SELECT MAX(EXPLAIN_TIME) FROM EXPLAIN_OBJECT) AND OBJECT_TYPE = 'IX'";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				String indexName = rs.getString("OBJECT_NAME");
				Index idx = new Index(indexName);
				idx.setNrLeafInitial(rs.getLong("NLEAF"));
				double size = Double.valueOf(rs.getString("NLEAF")) * 4096.0 / (1024.0*1024.0);
				idx.setSize(size);
				result.add(idx);
			}
			rs.close();
			closeStatement();
		} catch (SQLException e) {
			logger.error("Error in getUsedIndexes(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getUsedIndexes(): " + e.toString());
		}
		return result;
	}
	
	/**
	 * Generates all existing Indexes for the tables in tables.
	 * 
	 * @param tables
	 * @return all existing indexes
	 */
	public ListOfIndexes getAllExistingIndexes(ArrayList<String> tables) {
		logger.debug("Getting all Indexes for tables!");
		ListOfIndexes result = new ListOfIndexes();
		try {
			String query = "SELECT NAME, TBNAME, SYSTEM_REQUIRED, USER_DEFINED, COLNAMES, NLEAF FROM SYSIBM.SYSINDEXES";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				String tableName = rs.getString("TBNAME");
				if (tables.contains(tableName)) {
					String indexName = rs.getString("NAME");
					Index idx = new Index(indexName);
					String columns = rs.getString("COLNAMES");
					columns = columns.substring(1, columns.length());
					idx.setIndexTableAndColumn(tableName + "." + columns);
					idx.setNrLeafInitial(rs.getLong("NLEAF"));
					int userDefined = rs.getInt("USER_DEFINED");
					int systemRequired = rs.getInt("SYSTEM_REQUIRED");
					if (userDefined == 1) {
						idx.setUserDefined(true);
					}
					if (systemRequired == 1) {
						idx.setSystemRequired(true);
					}
					idx.setExists(true);
					result.add(idx);
				}
			}
			rs.close();
			closeStatement();
		} catch (SQLException e) {
			logger.error("Error in getAllExistingIndexes(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getAllExistingIndexes(): " + e.toString());
		}
		return result;
	}
	
	/**
	 * Updates the NLEAF for an user-defined existing index with indexName to newValue.
	 * 
	 * @param indexName
	 * @param newValue
	 */
	public void updateNLeaf(String indexName, long newValue) {
		logger.debug("Setting NLEAF to " + newValue + " for Index " + indexName + "!");
		String query = "UPDATE SYSIBM.SYSINDEXES SET NLEAF = " + newValue + " WHERE TBNAME = '" + indexName + "'";
		updateQuery(query);
	}
	
}
