package sqlindexwizard.dbconnection;

import java.sql.*;
import java.util.ArrayList;
import java.util.HashMap;

import org.apache.log4j.Logger;

import sqlindexwizard.clustering.TableElement;
import sqlindexwizard.filter.ItemWhereElement;

/**
 * Extracts selectivity from the database.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class SelectivityDB2 extends ConnectionDB2 {
	
	private static Logger logger = Logger.getLogger(SelectivityDB2.class);

	/**
	 * Constructor creates connection.
	 */
	public SelectivityDB2() {
		super();
	}
	
	/**
	 * Generates selectivity for an item.
	 * 
	 * @param it Item
	 * @param tab tableInfo
	 * @return selectivity
	 */
	public double getSelectivity(ItemWhereElement it, TableElement tab) {
		logger.debug("Getting selectivity for item " + it + "!");
		if (it.getType() == ItemWhereElement.TYPE_INTEGER) {
			return getSelectivityLong(it,tab);
		} else if (it.getType() == ItemWhereElement.TYPE_DATE){
			return getSelectivityLong(it,tab);
		} else if (it.getType() == ItemWhereElement.TYPE_TIME){
			return getSelectivityLong(it,tab);
		} else if (it.getType() == ItemWhereElement.TYPE_TIMESTAMP){
			return getSelectivityLong(it,tab);
		} else if (it.getType() == ItemWhereElement.TYPE_DOUBLE) {
			return getSelectivityDouble(it, tab);
		} else if (it.getType() == ItemWhereElement.TYPE_STRING) {
			return getSelectivityString(it,tab);
		} else if (it.getType() == ItemWhereElement.TYPE_UNKNOWN) {
			logger.error("UNKNOWN Type for Item " + it + "!");
			return 0.0;
		} else {
			logger.error("UNREACHABLE CODE REACHED!!! ITEM: " + it + "!");
			return 0.0;
		}
	}
	
	/**
	 * Generates Selectivity for an Item of type convertable to a long.
	 * Types: Integer, Time, Timestamp, Date
	 * 
	 * @param it Item
	 * @param tab tableInfo
	 * @return selectivity
	 */
	public double getSelectivityLong(ItemWhereElement it, TableElement tab) {
		double result = 0.0;
		try {
			String tableName = it.getTable().toUpperCase();
			String operator = it.getOperator();
			String column = it.getColumn().toUpperCase();
			byte type = it.getType();
			double count = 0.0;
			long tCard = tab.getTCard();
			if (tCard == 0) {
				logger.warn("Table " + tableName + "seems to be empty!");
			}
			long colCard = tab.getColCard(column);
			ArrayList<String> values = it.getValues();
			long value = getLongValue(type, values.get(0));
			long high2key = getLongValue(type, tab.getHigh2Key(column));
			long low2key = getLongValue(type, tab.getLow2Key(column));
			long range = high2key - low2key;
			if ((value - (range/colCard)) < low2key || (value + (range/colCard) > high2key)) {
				return 0.0;
			}
			if (operator.equals("=")) {
				HashMap<String,Integer> freqvals = getFrequentValues(tableName, column);
				if (freqvals.get(values.get(0)) != null) {
					count = freqvals.get(values.get(0));
				} else {
					String[][] quant = getQuantiles(tableName, column, tab.getNrQuantiles(column));
					for (int i = 0; i < quant.length; i++) {
						if (value <= getLongValue(type, quant[i][0]) || i == quant.length-1) {
							long thisRange = 0;
							long valuesInQuant = 0;
							if (i==0) {
								thisRange = getLongValue(type, quant[i][0]) - (getLongValue(type, tab.getLow2Key(column)) - ((range/colCard)));
								valuesInQuant = Long.valueOf(quant[i][1]);
							} else {
								thisRange = getLongValue(type, quant[i][0]) - getLongValue(type, quant[i-1][0]);
								valuesInQuant = Long.valueOf(quant[i][1]) - Long.valueOf(quant[i-1][1]);
							}
							count = 1.0*(valuesInQuant * range)/(colCard*thisRange);
							break;
						}
					}
				}
			} else if (operator.equals("<") || operator.equals(">") || operator.equals("<=") || operator.equals(">=")) {
				String[][] quant = getQuantiles(tableName, column, tab.getNrQuantiles(column));
				if (operator.equals("<")) {
					value = value - (range/colCard);
				} else if (operator.equals(">")) {
					value = value + (range/colCard);
				}
				for (int i = 0; i < quant.length; i++) {
					if (value <= getLongValue(type, quant[i][0]) || i == quant.length-1) {
						long thisRange = 0;
						long valuesInQuant = 0;
						long quantStartCount = 0;
						long quantStartValue = 0;
						if (i == 0) {
							quantStartCount = 0;
							quantStartValue = getLongValue(type, tab.getLow2Key(column)) - ((range/colCard));
							thisRange = getLongValue(type, quant[i][0]) - quantStartValue;
							valuesInQuant = Integer.valueOf(quant[i][1]);
						} else {
							quantStartCount = Integer.valueOf(quant[i-1][1]);
							quantStartValue = getLongValue(type, quant[i-1][0]);
							thisRange = getLongValue(type, quant[i][0]) - quantStartValue;
							valuesInQuant = Integer.valueOf(quant[i][1]) - quantStartCount;
						}
						count = quantStartCount + ((1.0*(value - quantStartValue)/thisRange)*valuesInQuant);
						if (operator.equals(">") || operator.equals(">=")) {
							count = tCard - count;
						}
						break;
					}
				}
			} else if (operator.equals("BETWEEN")) {
				ItemWhereElement it1 = new ItemWhereElement("<=",column,new String[]{values.get(0)});
				it1.setTable(tableName);
				it1.setType(it.getType());
				ItemWhereElement it2 = new ItemWhereElement("<=",column,new String[]{values.get(1)});
				it2.setTable(tableName);
				it2.setType(it.getType());
				return getSelectivityLong(it2,tab) - getSelectivityLong(it1, tab);
			} else if (operator.equals("IN")) {
				result = 0.0;
				for (int i = 0; i < values.size(); i++) {
					ItemWhereElement it1 = new ItemWhereElement("=",column,new String[]{values.get(i)});
					it1.setTable(tableName);
					it1.setType(it.getType());
					result += getSelectivityLong(it1,tab);
				}
				return result;
			} else if (operator.equals("!=")) {
				ItemWhereElement it1 = new ItemWhereElement("=",column,new String[]{values.get(0)});
				it1.setTable(tableName);
				it1.setType(it.getType());
				return (1.0 - getSelectivityLong(it1,tab));
			}
			if (count < 1) {
				count = 1;
			}
			result = (1.0*count)/tCard;
		} catch (Exception e) {
			logger.error("Exception in getSelectivityDouble(): " + e.getMessage());
		}
		if (result > 1.0) {
			result = 1.0;
		} else if (result < 0.0) {
			result = 0.0;
		}
		if (Double.isNaN(result)) {
			result = 1.0;
		}
		return result;
	}
	
	/**
	 * Generates Selectivity for an Item of type double.
	 * 
	 * @param it Item
	 * @param tab tableInfo
	 * @return selectivity
	 */
	public double getSelectivityDouble(ItemWhereElement it, TableElement tab) {
		double result = 0.0;
		try {
			String tableName = it.getTable().toUpperCase();
			String operator = it.getOperator();
			String column = it.getColumn().toUpperCase();
			double count = 0.0;
			long tCard = tab.getTCard();
			if (tCard == 0) {
				logger.warn("Table " + tableName + "seems to be empty!");
			}
			long colCard = tab.getColCard(column);
			ArrayList<String> values = it.getValues();
			double value = Double.valueOf(values.get(0));
			double high2key = Double.valueOf(tab.getHigh2Key(column));
			double low2key = Double.valueOf(tab.getLow2Key(column));
			double range = high2key - low2key;
			if ((value - (range/colCard)) < low2key || (value + (range/colCard) > high2key)) {
				return 0.0;
			}
			if (operator.equals("=")) {
				HashMap<String,Integer> freqvals = getFrequentValues(tableName, column);
				if (freqvals.get(values.get(0)) != null) {
					count = freqvals.get(values.get(0));
				} else {
					String[][] quant = getQuantiles(tableName, column, tab.getNrQuantiles(column));
					for (int i = 0; i < quant.length; i++) {
						if (value <= Double.valueOf(quant[i][0]) || i == quant.length-1) {
							double thisRange = 0;
							long valuesInQuant = 0;
							if (i==0) {
								thisRange = Double.valueOf(quant[i][0]) - (Double.valueOf(tab.getLow2Key(column)) - ((range/colCard)));
								valuesInQuant = Integer.valueOf(quant[i][1]);
							} else {
								thisRange = Double.valueOf(quant[i][0]) - Double.valueOf(quant[i-1][0]);
								valuesInQuant = Integer.valueOf(quant[i][1]) - Integer.valueOf(quant[i-1][1]);
							}
							count = 1.0*(valuesInQuant * range)/(colCard*thisRange);
							break;
						}
					}
				}
			} else if (operator.equals("<") || operator.equals(">") || operator.equals("<=") || operator.equals(">=")) {
				String[][] quant = getQuantiles(tableName, column, tab.getNrQuantiles(column));
				if (operator.equals("<")) {
					value = value - (range/colCard);
				} else if (operator.equals(">")) {
					value = value + (range/colCard);
				}
				for (int i = 0; i < quant.length; i++) {
					if (value <= Double.valueOf(quant[i][0]) || i == quant.length-1) {
						double thisRange = 0;
						long valuesInQuant = 0;
						long quantStartCount = 0;
						double quantStartValue = 0;
						if (i == 0) {
							quantStartCount = 0;
							quantStartValue = Double.valueOf(tab.getLow2Key(column)) - ((range/colCard));
							thisRange = Double.valueOf(quant[i][0]) - quantStartValue;
							valuesInQuant = Integer.valueOf(quant[i][1]);
						} else {
							quantStartCount = Integer.valueOf(quant[i-1][1]);
							quantStartValue = Double.valueOf(quant[i-1][0]);
							thisRange = Double.valueOf(quant[i][0]) - quantStartValue;
							valuesInQuant = Integer.valueOf(quant[i][1]) - quantStartCount;
						}
						count = quantStartCount + ((1.0*(value - quantStartValue)/thisRange)*valuesInQuant);
						if (operator.equals(">") || operator.equals(">=")) {
							count = tCard - count;
						}
						break;
					}
				}
			} else if (operator.equals("BETWEEN")) {
				ItemWhereElement it1 = new ItemWhereElement("<=",column,new String[]{values.get(0)});
				it1.setTable(tableName);
				it1.setType(it.getType());
				ItemWhereElement it2 = new ItemWhereElement("<=",column,new String[]{values.get(1)});
				it2.setTable(tableName);
				it2.setType(it.getType());
				return getSelectivityDouble(it2,tab) - getSelectivityDouble(it1, tab);
			} else if (operator.equals("IN")) {
				result = 0.0;
				for (int i = 0; i < values.size(); i++) {
					ItemWhereElement it1 = new ItemWhereElement("=",column,new String[]{values.get(i)});
					it1.setTable(tableName);
					it1.setType(it.getType());
					result += getSelectivityDouble(it1,tab);
				}
				return result;
			} else if (operator.equals("!=")) {
				ItemWhereElement it1 = new ItemWhereElement("=",column,new String[]{values.get(0)});
				it1.setTable(tableName);
				it1.setType(it.getType());
				return (1.0 - getSelectivityDouble(it1,tab));
			}
			if (count < 1) {
				count = 1;
			}
			result = (1.0*count)/tCard;
		} catch (Exception e) {
			logger.error("Exception in getSelectivityDouble(): " + e.getMessage());
		}
		if (result > 1.0) {
			result = 1.0;
		} else if (result < 0.0) {
			result = 0.0;
		}
		if (Double.isNaN(result)) {
			result = 1.0;
		}
		return result;
	}
	
	/**
	 * Generates Selectivity for an Item of type string (varchar, char).
	 * 
	 * @param it Item
	 * @param tab tableInfo
	 * @return selectivity
	 */
	public double getSelectivityString(ItemWhereElement it, TableElement tab) {
		//TODO
		return 0.0;
	}
	
	/**
	 * Getting frequent values for a column from the database.
	 * 
	 * @param tableName tableName
	 * @param column column
	 * @return hashmap value count
	 */
	public HashMap<String,Integer> getFrequentValues(String tableName, String column) {
		logger.debug("Getting frequent values for " + tableName + "." + column);
		HashMap<String,Integer> result = new HashMap<String,Integer>();
		try {
			String query = "SELECT COLVALUE, VALCOUNT FROM SYSIBM.SYSCOLDIST WHERE TYPE = 'F' AND NAME = '" + column.toUpperCase() + "'AND TBNAME = '" + tableName.toUpperCase() + "'";
			ResultSet rs = executeQuery(query);
			while (rs.next()) {
				result.put(rs.getString("COLVALUE"),Integer.valueOf(rs.getString("VALCOUNT")));
			}
			rs.close();
			closeStatement();
		} catch (SQLException e) {
			logger.error("Error in getFrequentValues(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getFrequentValues(): " + e.toString());
		}
		return result;	
	}
	
	/**
	 * Getting quantiles for a column from the database.
	 * 
	 * @param tableName tablename
	 * @param column column
	 * @param size nrOfQuantiles
	 * @return quantiles as a 2d string array (value, count)
	 */
	public String[][] getQuantiles(String tableName, String column, int size) {
		logger.debug("Getting quantiles values for " + tableName + "." + column);
		String[][] result = new String[size][2];
		try {
			String query = "SELECT SEQNO, COLVALUE, VALCOUNT FROM SYSIBM.SYSCOLDIST WHERE TYPE = 'Q' AND NAME = '" + column.toUpperCase() + "'AND TBNAME = '" + tableName.toUpperCase() + "' ORDER BY SEQNO ASC";
			ResultSet rs = executeQuery(query);
			int i = 0;
			while (rs.next()) {
				result[i][0] = rs.getString("COLVALUE");
				result[i][1] = rs.getString("VALCOUNT");
				i++;
			}
			rs.close();
			closeStatement();
		} catch (SQLException e) {
			logger.error("Error in getQuantiles(): " + e.toString());
		} catch (Exception e) {
			logger.error("Error in getQuantiles(): " + e.toString());
		}
		return result;
	}
	
	/**
	 * Calculates the long value of an element according to its type.
	 * Possible for: Integer, Time, Date, Timestamp
	 * 
	 * @param type type
	 * @param value value as String
	 * @return doubleValue
	 */
	@SuppressWarnings("finally")
	public static long getLongValue(byte type, String value) {
		try {
			if (type == ItemWhereElement.TYPE_INTEGER) {
				return Long.valueOf(value);
			} else if (type == ItemWhereElement.TYPE_TIME) {
				return Time.valueOf(value.substring(1, value.length()-1)).getTime();
			} else if (type == ItemWhereElement.TYPE_DATE) {
				return Date.valueOf(value.substring(1, value.length()-1)).getTime();
			} else if (type == ItemWhereElement.TYPE_TIMESTAMP) {
				return Timestamp.valueOf(value.substring(1, value.length()-1)).getTime();
			} else {
				logger.error("Not able to calculate LongValue of " + value + "!");
				return 0;
			}
		} catch (NumberFormatException e) {
			logger.error("Error in getDoubleValue(" + type + "," + value + "): " + e.getMessage());
		} finally {
			return 0;
		}
	}
	
	/**
	 * Calculates the double value of an element according to its type.
	 * Possible for: Integer, Time, Date, Timestamp, Double
	 * 
	 * @param type type
	 * @param value value as String
	 * @return doubleValue
	 */
	public static double getDoubleValue(byte type, String value) {
		if (type == ItemWhereElement.TYPE_DOUBLE) {
			double result = 0.0;
			try {
				result = Double.valueOf(value);
			} catch (NumberFormatException e) {
				logger.error("Error in getDoubleValue(" + type + "," + value + "): " + e.getMessage());
			}
			return result;
		} else {
			return 1.0*getLongValue(type, value);
		}
	}
	
}
