package sqlindexwizard.filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Vector;

import org.apache.log4j.Logger;

import Zql.ZExp;
import Zql.ZExpression;
import Zql.ZQuery;

/**
 * Representation of a where clause.
 * 
 * @author Benjamin Daeumlich
 *
 */
public abstract class WhereElement implements Cloneable {
	
	protected ListOfWhereElements elements = null;
	protected ArrayList<ItemWhereElement> itemList = new ArrayList<ItemWhereElement>();
	private boolean isValid = true;
	
	private static Logger logger = Logger.getLogger(WhereElement.class);
	
	/**
	 * Constructor generates an empty whereElement.
	 */
	public WhereElement() {
		elements = new ListOfWhereElements();
	}
	
	/**
	 * Constructor generates a WhereElement for a List of WhereElements.
	 * 
	 * @param elements
	 */
	public WhereElement(ListOfWhereElements elements) {
		this.elements = elements;
	}
	
	/**
	 * Parses the WhereClause.
	 * 
	 * @param where whereclause
	 * @return whereElement
	 */
	@SuppressWarnings("unchecked")
	public static WhereElement parseWhereClause(ZExp where) {
		WhereElement whereelem = null;
		if (where instanceof ZExpression) {
			String operator = ((ZExpression) where).getOperator();
			Vector<ZExp> newelems = ((ZExpression) where).getOperands();
			if (operator.equals("AND")) {
				whereelem = new AndWhereElement();
				for (int i = 0; i < newelems.size(); i++) {
					whereelem.add(parseWhereClause(newelems.get(i)));
				}
			} else if (operator.equals("OR")) {
				whereelem = new OrWhereElement();
				for (int i = 0; i < newelems.size(); i++) {
					whereelem.add(parseWhereClause(newelems.get(i)));
				}
			} else if (operator.equals("NOT")) {
				whereelem = new NotWhereElement();
				whereelem.add(parseWhereClause(newelems.get(0)));
			} else {
				whereelem = new ItemWhereElement(operator, newelems);
			}
		} else if (where instanceof ZQuery) {
			logger.info("No Subquerys supported!");
			return null;
		} else {
			whereelem = new EmptyWhereElement();
		}
		return whereelem;
	}
	
	/**
	 * Converts this filter into CNF. Used Rules
	 * 
	 * Part 1 (nothing to do):
	 * 
	 * 		a <-> b = (NOT(a) or b) and (a or NOT(b))
	 * 		a  -> b = NOT(a) or b
	 * 
	 * Part 2:
	 * 
	 * 		NOT(a or b) = NOT(a) and NOT(b)
	 * 		NOT(a and b) = NOT(a) or NOT(b)
	 *      NOT(NOT(a)) = a
	 * 
	 * Part 3:
	 * 
	 * 		a or (b and c) = (a or b) and (a or c)
	 * 
	 * Part 4 (simplify) (not implemented):
	 * 
	 * 		a and  a = a
	 * 		a and NOT(a) = 0
	 * 		a or  a = a
	 * 		a or NOT(a) = 1
	 * 		a and (a or b) = a
	 * 
	 * @return WhereElement in CNF
	 */
	public abstract WhereElement toCNF();
	
	/**
	 * Converts this filter into DNF. Used Rules
	 * 
	 * Part 1 (nothing to do):
	 * 
	 * 		a <-> b = (a and b) or (NOT(a) and NOT(b))
	 * 		a  -> b = NOT(a) or b
	 * 
	 * Part 2:
	 * 
	 * 		NOT(a or b) = NOT(a) and NOT(b)
	 * 		NOT(a and b) = NOT(a) or NOT(b)
	 *      NOT(NOT(a)) = a
	 * 
	 * Part 3:
	 * 
	 * 		a and (b or c) = (a and b) or (a and c)
	 * 
	 * Part 4 (simplify) (not implemented):
	 * 
	 * 		a and  a = a
	 * 		a and NOT(a) = 0
	 * 		a or  a = a
	 * 		a or NOT(a) = 1
	 * 		a or (a and b) = a
	 * 
	 * @return WhereElement in DNF
	 */
	public abstract WhereElement toDNF();
	
	/**
	 * Simplifies WhereElement. Not implemented yet.
	 * 
	 * @return simplified whereElement
	 */
	public abstract WhereElement simplify();
	
	/**
	 * Removes parentheses not recursively.
	 * 
	 * @return whereElement without parentheses
	 */
	public abstract WhereElement removeParThisLevel();
	
	/**
	 * Removes parentheses from this WhereElement recursively.
	 * 
	 * @return WhereElement without parentheses
	 */
	public WhereElement removePar() {
		if (elements == null || elements.size() == 0) {
			return this.removeParThisLevel();
		}
		WhereElement where;
		try {
			where = this.getClass().newInstance();
		} catch (Exception e) {
			throw new RuntimeException("Cannot create WhereElement");
		}
		// 1st: subfilters
		for (WhereElement ft : this.elements) {
			where.add(ft.removePar());
		}
		// 2nd: this level
		return where.removeParThisLevel();
	}
	
	/**
	 * Tests for equality of two Objects.
	 * 
	 * @return true if they are equal
	 */
	public boolean equals(Object o) {
		if (o instanceof WhereElement) {
			return equals((WhereElement) o);
		} else {
			return false;
		}
	}

	/**
	 * Tests for equality of two Objects. Element-Lists have to be the
	 * same.
	 * 
	 * @return true if they are equal
	 */
	public boolean equals(WhereElement it) {
		if (this == it) return true;
		if (it == null) return false;
		if (getClass() != it.getClass()) return false;
		if (elements == null) {
			if (it.elements != null) {
				return false;
			}
		} else if (elements.size() == 0 && it.elements.size() == 0) {
			if (it instanceof ItemWhereElement) {
				return ((ItemWhereElement)it).equals((ItemWhereElement)this);
			} else if (it instanceof EmptyWhereElement) {
				return true;
			} else if (it instanceof ContradictionWhereElement) {
				return true;
			} else if (it instanceof TautologyWhereElement) {
				return true;
			}
		} else if (!elements.equals(it.elements)) {
			return false;
		}
		return true;
	}
	
	/**
	 * Tests for equality of two Objects for Sim-Calculation. Element-Lists have to be the
	 * same.
	 * 
	 * @return true if they are equal
	 */
	public boolean equalsSim(WhereElement it) {
		if (this == it) return true;
		if (it == null) return false;
		if (getClass() != it.getClass()) return false;
		if (elements == null) {
			if (it.elements != null) {
				return false;
			}
		} else if (elements.size() == 0 && it.elements.size() == 0) {
			if (it instanceof ItemWhereElement) {
				return ((ItemWhereElement)it).equalsSim((ItemWhereElement)this);
			} else if (it instanceof EmptyWhereElement) {
				return true;
			} else if (it instanceof ContradictionWhereElement) {
				return true;
			} else if (it instanceof TautologyWhereElement) {
				return true;
			}
		} else if (!elements.equalsSim(it.elements)) {
			return false;
		}
		return true;
	}
	
	/**
	 * Clone this WhereElement.
	 * 
	 * @return clone 
	 */
	public WhereElement clone() {
		try {
			WhereElement parent = this.getClass().newInstance();
			if (elements != null) {
				for (WhereElement child : elements) {
					parent.add(child.clone());
				}
			}
			return parent;
		} catch (Exception e) {
			throw new RuntimeException("Cannot clone WhereElement");
		}
	}
	
	/**
	 * Adding an element to the whereelement.
	 * 
	 * @param element el to add
	 * @return true if added
	 */
	public boolean add (WhereElement element) {
		return this.elements.add(element);
	}
	
	/**
	 * Adding a collection of whereelements to this whereelement.
	 * 
	 * @param where whereelements to add
	 * @return true if added
	 */
	public boolean addAll(Collection<? extends WhereElement> where) {
		return elements.addAll(where);
	}
	
	/**
	 * Getting site of the WhereElement.
	 * 
	 * @return size
	 */
	public int size() {
		return elements.size();
	}
	
	/**
	 * Getting element at index i.
	 * 
	 * @param index
	 * @return WhereElement at index
	 */
	public WhereElement get(int index) {
		return elements.get(index);
	}
	
	/**
	 * Getting all items of the WhereElement.
	 * 
	 * @return itemList
	 */
	public ArrayList<ItemWhereElement> getItemList() {
		ArrayList<ItemWhereElement> result = new ArrayList<ItemWhereElement>();
		if (this instanceof ItemWhereElement) {
			result.add((ItemWhereElement)this);
		}
		for (WhereElement element: elements) {
			if (element instanceof ItemWhereElement) {
				result.addAll(element.getItemList());
			} else if (element instanceof OrWhereElement) {
				result.addAll(element.getItemList());
			} else if (element instanceof AndWhereElement) {
				result.addAll(element.getItemList());
			} else if (element instanceof NotWhereElement) {
				ListOfWhereElements temp = element.elements;
				for (int i = 0; i < temp.size(); i++) {
					if (temp.get(i) instanceof ItemWhereElement) {
					((ItemWhereElement)temp.get(i)).setNot(true);
					}
				}
				result.addAll(element.getItemList());
			}
		}
		return result;
	}
	
	/**
	 * Is the WhereElement valid?
	 * 
	 * @return the isValid
	 */
	public boolean isValid() {
		return isValid;
	}

	/**
	 * Setting isValid.
	 * 
	 * @param isValid the isValid to set
	 */
	public void setValid(boolean isValid) {
		this.isValid = isValid;
	}

	/**
	 * String representation of the where clause.
	 * 
	 * @return string representation
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer();
		for (WhereElement where : elements) {
			buffer.append(where.toString());
		}
		return buffer.toString();
	}

}
