package sqlindexwizard.indexadvisor;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

//import at.repher.console.Progress;

import sqlindexwizard.application.Input;
import sqlindexwizard.dbconnection.IndexDB2;
import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.filter.JoinElement;
import sqlindexwizard.filter.SelectElement;
import sqlindexwizard.util.Config;
import sqlindexwizard.util.ProgressBar;
import sqlindexwizard.workload.ListOfQueries;
import sqlindexwizard.workload.Workload;

/**
 * IndexAdvisor gets useful indexes from the database and selects the most profitable ones fitting
 * a space constraint.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class IndexAdvisor {
	
	private ListOfIndexes recommendedIndexes = new ListOfIndexes();
	private ListOfIndexes possibleIndexes = new ListOfIndexes();
	private ListOfIndexes allExistingIndexes = new ListOfIndexes();
	private ListOfIndexes deleteableIndexes = new ListOfIndexes();
	
	private static Logger logger = Logger.getLogger(IndexAdvisor.class);
	
	private Workload workload;

	/**
	 * Constructor sets local variable workload.
	 * 
	 * @param workload Workload with queries
	 */
	public IndexAdvisor(Workload workload) {
		this.workload = workload;
	}
	
	/**
	 * Fills the ADVISE_INDEX - table with entries for each table-column of the workload.
	 * 
	 */
	private void fillAdviseIndex() {
		logger.info("Filling ADVISE_INDEX table!");
		IndexDB2 idx = new IndexDB2();
		ArrayList<SelectElement> otherList = Workload.getOtherList();
		ArrayList<ItemWhereElement> itemList = Workload.getItemList();
		ArrayList<JoinElement> joinList = Workload.getJoinList();
		ArrayList<String> columns = new ArrayList<String>();
		for (int i = 0; i < itemList.size(); i++) {
			ItemWhereElement temp = itemList.get(i);
			if (!columns.contains(temp.getTable() + "." + temp.getColumn())) {
				columns.add(temp.getTable() + "." + temp.getColumn());
			}
		}
		for (int i = 0; i < joinList.size(); i++) {
			JoinElement temp = joinList.get(i);
			if (!columns.contains(temp.getTableLeft() + "." + temp.getColumnLeft())) {
				columns.add(temp.getTableLeft() + "." + temp.getColumnLeft());
			}
			if (!columns.contains(temp.getTableRight() + "." + temp.getColumnRight())) {
				columns.add(temp.getTableRight() + "." + temp.getColumnRight());
			}
		}
		for (int i = 0; i < otherList.size(); i++) {
			SelectElement temp = otherList.get(i);
			if (!columns.contains(temp.getTable() + "." + temp.getColumn()) && !temp.isWildcard()) {
				columns.add(temp.getTable() + "." + temp.getColumn());
			}
		}
		idx.cleanAdviseIndex();
		idx.setExplainModeRecommend();
		for (int i = 0; i < columns.size(); i++) {
			StringTokenizer stok = new StringTokenizer(columns.get(i), ".");
			String table = stok.nextToken();
			String column = stok.nextToken();
			idx.createAdviseIndexEntry(column, table);
		}
		idx.setExplainModeNo();
		idx.finalize();
		logger.info("Finished filling ADVISE_INDEX table!");
	}
	
	/**
	 * Takes all possible Indexes for the workload form the ADVISE_INDEX table.
	 * 
	 */
	private void createPossibleIndexes() {
		logger.info("Creating possible Indexes for Workload!");
		IndexDB2 idx = new IndexDB2();
		possibleIndexes = idx.getPossibleIndexes();
		idx.finalize();
		logger.info("Finished creating possible Indexes for Workload!");
	}
	
	/**
	 * Executes the queries and collects used virtual indexes and costs.
	 * 
	 * @param queryList ListOfQueries
	 * @return indexList
	 */
	private ListOfIndexes generateIndexesAndBenefits(ListOfQueries queryList) {
		logger.info("Generating Indexes and Benefits!");
		ListOfIndexes indexList = new ListOfIndexes();
		ProgressBar progress = new ProgressBar();
		for (int i = 0; i < queryList.size(); i++) {
			IndexDB2 idx = new IndexDB2();
			double costInitial = 0.0;
			double costVirtual = 0.0;
			ListOfIndexes indexes = new ListOfIndexes();
			String query = queryList.get(i).toString();
			query = query.substring(0,query.length()-1);
			idx.setExplainModeEvaluate();
			costVirtual = idx.getQueryCost(query);
			if (costVirtual == -9999999999.0) {
				progress.setProgress((i+1),queryList.size());
				continue;
			}
			indexes = idx.getUsedIndexes();
			ListOfIndexes userDefinedExistingIndexes = new ListOfIndexes();
			int nrNotSysReq = indexes.size();
			for (int j = 0; j < indexes.size(); j++) {
				Index index;
				int indexOf = possibleIndexes.indexOf(indexes.get(j));
				if (indexOf >= 0) {
					index = possibleIndexes.get(indexOf);
					index.setSize(indexes.get(j).getSize());
				} else {
					index = idx.getIndex(indexes.get(j).getIndexName());
					index.setSize(indexes.get(j).getSize());
				}
				indexes.set(j, index);
				if (index.isSystemRequired()) {
					nrNotSysReq--;
				}
				if (index.isExists() && index.isUserDefined()) {
					userDefinedExistingIndexes.add(index);
					idx.updateNLeaf(index.getIndexName(), 1111111);
				}
			}
			idx.setExplainModeExplain();
			costInitial = idx.getQueryCost(query);
			for (int j = 0; j < userDefinedExistingIndexes.size(); j++) {
				idx.updateNLeaf(userDefinedExistingIndexes.get(j).getIndexName(), userDefinedExistingIndexes.get(j).getNrLeafInitial());
			}
			for (int j = 0; j < indexes.size(); j++) {
				Index index = indexes.get(j);
				double benefit = 0.0;
				if (!index.isSystemRequired()) {
					// modify-queries: other benefit calculation, not implemented
					benefit = (costInitial - costVirtual)*queryList.get(i).getCount()/nrNotSysReq;
					if (benefit < 0) {
						break;
					}
				}
				if (!indexList.contains(index)) {
					index.setBenefit(benefit);
					indexList.add(index);
				} else {
					index = indexList.get(indexList.indexOf(index));
					index.increaseBenefit(benefit);
				}
			}
			idx.finalize();
			progress.setProgress((i+1),queryList.size());
		}
		progress.setProgress(queryList.size(),queryList.size());
		// finally remove system-required indexes
		ListOfIndexes indexListFinal = new ListOfIndexes();
		for (int i = 0; i < indexList.size(); i++) {
			if (!indexList.get(i).isSystemRequired()) {
				indexListFinal.add(indexList.get(i));
			}
		}
		logger.info("Finished generating Indexes and Benefits!");
		return indexListFinal;
	}
	
	/**
	 * Generates all existing Indexes for the referenced tables in the Workload.
	 * 
	 */
	private void generateAllExistingIndexes() {
		logger.info("Generating all Indexes!");
		IndexDB2 idx = new IndexDB2();
		ArrayList<String> tables = new ArrayList<String>();
		tables.addAll(Workload.getMapOfTableElements().keySet());
		allExistingIndexes = idx.getAllExistingIndexes(tables);
		idx.finalize();
		logger.info("Finished generating all Indexes!");
	}
	
	/**
	 * Generates deleteable Indexes.
	 * 
	 */
	private void generateDeleteableIndexes() {
		logger.info("Generating deleteable Indexes!");
		for (int i = 0; i < allExistingIndexes.size(); i++) {
			Index index = allExistingIndexes.get(i);
			if (index.isUserDefined()) {
				if (!recommendedIndexes.contains(index)) {
					deleteableIndexes.add(index);
				}
			}
		}
		logger.info("Finished generating deleteable Indexes!");
	}
	
	/**
	 * Generates the indexes for the queries of the workload and runs greedy-algo.
	 * 
	 */
	public void goIndexAdvisor() {
		logger.info("Starting IndexAdvisor!");
		fillAdviseIndex();
		createPossibleIndexes();
		generateAllExistingIndexes();
		ListOfQueries input = workload.getSelectQueryList();
		// preparation for modify-queries
//		input.addAll(workload.getUpdateQueryList());
//		input.addAll(workload.getDeleteQueryList());
//		input.addAll(workload.getInsertQueryList());
		input.addAll(workload.getNotAddedQueryList());
		ListOfIndexes allIndexes = generateIndexesAndBenefits(input);
		if (Input.greedy) {
			recommendedIndexes = KnapsackAlgorithm.greedy(allIndexes, Config.getPropertyLong("knapsack.indexSpace"));
		} else if (Input.dynProg) {
			recommendedIndexes = KnapsackAlgorithm.dynamicProgramming(allIndexes, Config.getPropertyLong("knapsack.indexSpace"));
		} else if (Input.dynProgEpsilon) {
			if (Input.epsilon <= 0.0) {
				logger.error("Invalid value for epsilon: " + Input.epsilon + "!");
			}
			recommendedIndexes = KnapsackAlgorithm.dynamicProgrammingEpsilon(allIndexes, Config.getPropertyLong("knapsack.indexSpace"), Input.epsilon);
		} else {
			logger.error("Unknown knapsack algorithm");
		}
		generateDeleteableIndexes();
		generateOutputFile();
		logger.info("Finished IndexAdvisor!");
	}

	/**
	 * Getting recommended indexes.
	 * 
	 * @return recommended Indexes
	 */
	public ListOfIndexes getRecommendedIndexes() {
		return recommendedIndexes;
	}
	
	/**
	 * Generates recommendation-output-file with creation and drop commands for indexes.
	 */
	private void generateOutputFile() {
		try{
			File f = new File(Config.getPropertyString("general.outputFile","./result.txt"));
			try {
				File dir = new File(f.getParent());
				dir.mkdirs();
			} catch (Exception e) {
				// nothing todo, directory not created
			}
			logger.info("Generating Output into file " + f.getName() + ".");
			FileWriter out = new FileWriter(f);
			out.write("CONNECT TO " + Config.getPropertyString("db2.database") + ";\n");
			for (int i = 0; i < recommendedIndexes.size(); i++) {
				out.write(recommendedIndexes.get(i).getIndexCreationText() + ";\n");
			}
			for (int i = 0; i < deleteableIndexes.size(); i++) {
				out.write("DROP INDEX " + deleteableIndexes.get(i).getIndexName() + ";\n");
			}
			out.write("COMMIT WORK;\n");
			out.write("CONNECT RESET;\n");
			out.close();
		} catch (IOException e) {
			logger.error("IOException in generateOutputFile(): " + e.toString());
		}
	}
	
	/**
	 * String representation of Index-Advisor.
	 * 
	 * @return String representation
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		if (Input.greedy) {
			sb.append("Algorithm:\tGreedy\n");
		} else if (Input.dynProg) {
			sb.append("Algorithm:\tdynamic programming\n");
		} else if (Input.dynProgEpsilon) {
			sb.append("Algorithm:\tdynamic programming (epsilon = " + Input.epsilon + ")\n");
		}
		sb.append("Space-Constraint:\t" + Config.getPropertyDouble("knapsack.indexSpace") + "\n");
		sb.append("Index-Space:\t" + recommendedIndexes.getIndexSize() + "\n");
		sb.append("Total Benefit:\t" + recommendedIndexes.getBenefit() + "\n");
		sb.append("\nRecommended Indexes:\n\n");
		for (int i = 0; i < recommendedIndexes.size(); i++) {
			sb.append("\t" + recommendedIndexes.get(i).toStringName() + "\n");
		}
		sb.append("\nDeleteable Indexes:\n\n");
		for (int i = 0; i < deleteableIndexes.size(); i++) {
			sb.append("\t" + deleteableIndexes.get(i).toStringName() + "\n");
		}
		return sb.toString();
	}
	
}
