package sqlindexwizard.indexadvisor;

import java.util.Arrays;
import java.util.Comparator;

import org.apache.log4j.Logger;

/**
 * Knapsack-Algorithms are implemented here.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class KnapsackAlgorithm {
	
	private static Logger logger = Logger.getLogger(KnapsackAlgorithm.class);

	/**
	 * Implementation of the greedy algorithm.
	 * 
	 * @param indexes indexList
	 * @param space constraint
	 * @return indexes fitting space constraint
	 */
	public static ListOfIndexes greedy(ListOfIndexes indexes, long space) {
		logger.info("Starting greedy algorithm for Indexes!");
		ListOfIndexes result = new ListOfIndexes();
		if (indexes.size() == 0) {
			return result;
		}
		Index[] indexArray = indexes.toArray();
		Arrays.sort(indexArray, new Comparator<Index>() {
			public int compare(Index idx1, Index idx2) {
				if (idx1.getSize() == 0) {
					if (idx2.getSize() == 0) {
						return idx1.getBenefit() >= idx2.getBenefit() ?
								-1 : 1;
					} else {
						if (idx1.getBenefit() > 0) {
							return -1;
						} else {
							if (idx2.getBenefit() > 0) {
								return 1;
							} else {
								return idx1.getBenefit() >= idx2.getBenefit() ?
										-1 : 1;
							}
						}
					}
				} else {
					if (idx2.getSize() == 0) {
						if (idx2.getBenefit() > 0) {
							return 1;
						} else {
							if (idx1.getBenefit() > 0) {
								return -1;
							} else {
								return idx1.getBenefit() >= idx2.getBenefit() ?
										-1 : 1;
							}
						}
					} else {
						double value1 = idx1.getBenefit() / idx1.getSize();
						double value2 = idx2.getBenefit() / idx2.getSize();
						return value1 >= value2 ? -1 : 1;
					}
				}
			}
		});
		double knapsackSize = 0.0;
		double knapsackBenefit = 0.0;
		Index mostProfitableIdx = indexArray.length > 0 ? indexArray[0] : null;
		for (int i = 0; i < indexArray.length; i++) {
			if ((indexArray[i].getSize() + knapsackSize) <= space && (indexArray[i].getBenefit() > 0)) {
				result.add(indexArray[i]);
				knapsackSize += indexArray[i].getSize();
				knapsackBenefit += indexArray[i].getBenefit();
			}
			if (indexArray[i].getBenefit() > mostProfitableIdx.getBenefit() && indexArray[i].getSize() <= space) {
				mostProfitableIdx = indexArray[i];
			}
		}
		if (knapsackBenefit < mostProfitableIdx.getBenefit()) {
			result.clear();
			result.add(mostProfitableIdx);
		}
		logger.info("Finished greedy algorithm for Indexes!");
		return result;
	}
	
	/**
	 * Implementation of the dynamic programming algorithm.
	 * 
	 * @param indexes indexList
	 * @param space constraint
	 * @return indexes fitting space constraint
	 */
	public static ListOfIndexes dynamicProgramming(ListOfIndexes indexes, long space) {
		logger.info("Starting dynamic programming algorithm for Indexes!");
		ListOfIndexes result = new ListOfIndexes();
		if (indexes.size() == 0) {
			return result;
		}
		ListOfIndexes usefulIndexes = new ListOfIndexes();
		int benefitResult = 0; //Resulting benefit
		long benefitSumLong = 0; // sum of benefits
		long sizeSum = 0; // sum of sizes
		long maxBenefit = 0; // maximal Benefit
		// getting useful indexes with size <= space, generating totalBenefit, maxBenefit and sizeSum
		for (int i = 0; i < indexes.size(); i++) {
			Index temp = indexes.get(i);
			if (((long)temp.getSize() +1) <= space) {
				benefitSumLong += (long) Math.round(temp.getBenefit());
				sizeSum += ((long)temp.getSize() +1);
				usefulIndexes.add(temp);
				if ((long) Math.round(temp.getBenefit()) > maxBenefit) {
					maxBenefit = ((long) Math.round(temp.getBenefit()));
				}
				
			}
		}
		// case no useful indexes
		if (usefulIndexes.size() == 0) {
			return result;
		}
		// case all useful indexes fit into knapsack
		if (sizeSum < space) {
			return usefulIndexes;
		}
		int benefitSum = 0;
		// benefitSum hast do be an integer due to limitation of
		// array indexes to int-values
		if (benefitSumLong < Integer.MAX_VALUE) {
			benefitSum = (int)benefitSumLong;
		} else {
			double factor = 1.0*Integer.MAX_VALUE/(1000*(benefitSumLong+1.0));
			return dynamicProgrammingScaled(indexes,space,factor);
		}
		// generate arrays for dynamic programming, first for sizes and second for indexes
		// (size(i,p) = 
		//				(1) 0, if p <= 0
		//				(2) infinity if i = 0, p > 0
		//				(3) min (size(i-1,p), size_i + size(i-1, p-p_i)) otherwise
		int indexesSize = usefulIndexes.size();
		long[][] sizeTable = new long[indexesSize][];
		Index[][] recTable = new Index[indexesSize][];
		// step (1)
		for (int i = 0; i < indexesSize; i++) {
			sizeTable[i] = new long[benefitSum + 1];
			sizeTable[i][0] = 0;
			recTable[i] = new Index[benefitSum + 1];
			recTable[i][0] = null;
		}
		// step (2)
		for (int p = 1; p < benefitSum + 1; p++) {
			sizeTable[0][p] = sizeSum + 1;	// dummy for infinity
		}
		// set size for first index (i = 1)
		if (usefulIndexes.get(0).getBenefit() >= 0) {
			sizeTable[0][(int) Math.round(usefulIndexes.get(0).getBenefit())] = ((long)usefulIndexes.get(0).getSize() + 1);
			recTable[0][(int) Math.round(usefulIndexes.get(0).getBenefit())] = usefulIndexes.get(0);
		}
		int sizeTabGreater = 0;
		int maxBenefitLine = benefitSum;
		// step (3)
		for (int p = 0; p < benefitSum + 1; p++) {	// p = 0 for indexes with p = 0
			for (int i = 1; i < indexesSize; i++) {
				if ((p - (int) Math.round(usefulIndexes.get(i).getBenefit()) >= 0)	// valid entry
						&& (sizeTable[i - 1][p - (int) Math.round(usefulIndexes.get(i).getBenefit())] + ((long)usefulIndexes.get(i).getSize() +1) <= sizeTable[i - 1][p])) {
					// min is size_i + size(i-1, p-p_i)
					sizeTable[i][p] = sizeTable[i - 1][p - (int) Math.round(usefulIndexes.get(i).getBenefit())] + ((long)usefulIndexes.get(i).getSize() +1);
					recTable[i][p] = usefulIndexes.get(i);
				} else {
					// min is size(i-1,p), 
					sizeTable[i][p] = sizeTable[i - 1][p];
					recTable[i][p] = recTable[i - 1][p];
				}
			}
			// check if finished
			if (sizeTable[indexesSize - 1][p] <= space) {
				// last line still less equal space -> continue
				sizeTabGreater = 0;
				continue;
			} else {	// size > space
				if (sizeTabGreater == 0) {
					// first time
					sizeTabGreater = p;
				} else {
					// size was greater than space in line sizeTabGreater
					if (p - sizeTabGreater >= maxBenefit) {
						// found maximal Line
						maxBenefitLine = p;
						break;
					}
				}
			}
		}
		// finding optimum
		for (int p = maxBenefitLine; p > 0; p--) {
			if (sizeTable[indexesSize - 1][p] <= space) {
				benefitResult = p;
				break;
			}
		}
		// filling knapsack
		int indexPos = indexesSize - 1;
		int benefitPos = benefitResult;
		while (indexPos >= 0 && benefitPos >= 0 && recTable[indexPos][benefitPos] != null) {
			if (!result.contains(recTable[indexPos][benefitPos])) {
				result.add(recTable[indexPos][benefitPos]);
			}
			// set new values
			if (indexPos - 1 >= 0
					&& (benefitPos - (long) Math.round(usefulIndexes.get(indexPos).getBenefit()) >= 0)	// table entry is defined
					&& (sizeTable[indexPos - 1][benefitPos - (int) Math.round(usefulIndexes.get(indexPos).getBenefit())] + ((long)usefulIndexes.get(indexPos).getSize() +1) <= sizeTable[indexPos - 1][benefitPos])) {
				benefitPos -= (long) Math.round(usefulIndexes.get(indexPos).getBenefit());
			}
			indexPos--;
		}
		logger.info("Finished dynamic programming algorithm for Indexes!");
		return result;
	}
	
	/**
	 * Implementation of the dynamic programming algorithm with scaling, needed when
	 * sum of benefits are greater than MAX_INTEGER.
	 * 
	 * @param indexes indexList
	 * @param space constraint
	 * @return indexes fitting space constraint
	 */
	public static ListOfIndexes dynamicProgrammingScaled(ListOfIndexes indexes, long space, double factor) {
		logger.info("Starting scaled dynamic programming algorithm for Indexes!");
		for (int i = 0; i < indexes.size(); i++) {
			double benefitNew = indexes.get(i).getBenefit()*factor;
			if ((benefitNew < 1.0 && indexes.get(i).getBenefit() > 0.0)) {
				indexes.get(i).setBenefit(1.0);
			} else if (benefitNew <= 0.0) {
				indexes.get(i).setBenefit(0.0);
			} else {
				indexes.get(i).setBenefit(benefitNew);
			}
		}
		ListOfIndexes recommendedIndexes = dynamicProgramming(indexes,space);
		for (int i = 0; i < recommendedIndexes.size(); i++) {
			recommendedIndexes.get(i).setBenefit(1.0*recommendedIndexes.get(i).getBenefit()/factor);
		}
		logger.info("Finished scaled dynamic programming algorithm for Indexes!");
		return recommendedIndexes;
	}
	
	/**
	 * Implementation of the dynamic programming algorithm with scaling, needed if runtime
	 * increase needed.
	 * 
	 * @param indexes indexList
	 * @param space constraint
	 * @param epsilon
	 * @return indexes fitting space constraint
	 */
	public static ListOfIndexes dynamicProgrammingEpsilon(ListOfIndexes indexes, long space, double epsilon) {
		logger.info("Starting dynamic programming algorithm for Indexes with epsilon (" + epsilon + ")!");
		// calculation pMax
		double pMax = 0.0;
		for (int i = 0; i < indexes.size(); i++) {
			if (indexes.get(i).getBenefit() > pMax) {
				pMax = indexes.get(i).getBenefit();
			}
		}
		// calculation factor an set net benefits
		double factor = indexes.size()/(epsilon*pMax);
		if (factor >= 1) {
			logger.info("Finished dynamic programming algorithm for Indexes with epsilon without scaling!");
			return indexes;
		}
		for (int i = 0; i < indexes.size(); i++) {
			double benefitNew = indexes.get(i).getBenefit()*factor;
			if ((benefitNew < 1.0 && indexes.get(i).getBenefit() > 0.0)) {
				indexes.get(i).setBenefit(1.0);
			} else if (benefitNew <= 0.0) {
				indexes.get(i).setBenefit(0.0);
			} else {
				indexes.get(i).setBenefit(benefitNew);
			}
		}
		ListOfIndexes recommendedIndexes = dynamicProgramming(indexes,space);
		// reset benefits
		for (int i = 0; i < recommendedIndexes.size(); i++) {
			recommendedIndexes.get(i).setBenefit(1.0*recommendedIndexes.get(i).getBenefit()/factor);
		}
		logger.info("Finished dynamic programming algorithm for Indexes with epsilon!");
		return recommendedIndexes;
	}
	
}
