package sqlindexwizard.querygen;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * This class generates simple queries randomly. Getting information about the
 * db-schema from an input file.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class QueryGen {

	private ArrayList<Table> tables = new ArrayList<Table>();
	
	private String[] operators = {"=", ">=", "<="};
	private String[] logical = {"AND", "OR"};
	private String[] aggregate = {"MAX", "MIN", "AVG", "COUNT", "SUM"};
	
	// probabilities
	private double oneTable = 0.9; // one table ? 0.7
	private double wildCard = 0.5; // select * ? 0.5
	private double where = 1.0; // where-clause ? 0.7
	private double join = 0.1; // join in where-clause ? 0.1
	private double groupBy = 0.1; // group by-clause ? 0.1
	private double orderBy = 0.1; // order by-clause ? 0.1
	private double aggragateProb = 0.1; // aggregate for selectitem ? 0.1
	private double not = 0.05; // not for whereelement 0.05
	private int maxSelect = 3; // max nr of selectitems 3
	private int minWhere = 3; // min nr of whereitems 2
	private int maxWhere = 6; // max nr of whereitems 5
	
	// input and output options
	private String inputFileName = "./data/qGenInput.txt";
	private String outputFileName = "./data/qGenOutput100.txt";
	private int nrSimilarQueries = 4; // 4
	private int nrQueries = 100;
	
	private static int generatedQueries = 0;
	
	/**
	 * Reading in tables, columns and values from an input-file.
	 * 
	 */
	@SuppressWarnings("deprecation")
	public void readTableInformation() {
		File f = new File(inputFileName);
		try {
			DataInputStream s = new DataInputStream(new FileInputStream(f));
			String line = s.readLine();
			while (line != null) {
				processTableString(line);
				line = s.readLine();
			}
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found: " + f.toString());
		} catch (IOException e) {
			System.out.println("IOException!" + e.toString());
		}
	}
	
	/**
	 * Processing tableString.
	 * Format: tableName "," [columnName "|" [value ";"]+ "+"]+
	 * 
	 * @param s tableString
	 */
	public void processTableString(String s) {
		StringTokenizer st = new StringTokenizer(s,",");
		String tableName = st.nextToken();
		Table table = new Table(tableName);
		StringTokenizer st2 = new StringTokenizer(st.nextToken(),"+");
		while (st2.hasMoreTokens()) { 
			StringTokenizer st3 = new StringTokenizer(st2.nextToken(),"|");
			String columnName = st3.nextToken();
			Column column = new Column(columnName);
			StringTokenizer st4 = new StringTokenizer(st3.nextToken(),";");
			while (st4.hasMoreTokens()) {
				String valueString = st4.nextToken();
				column.addValue(valueString);
			}
			table.addColumn(column);
		}
		tables.add(table);
	}
	
	/**
	 * Generating query randomly. Generates nrSimilar similar queries (diffent values in
	 * where-clause).
	 * 
	 * @param nrSimilar
	 * @return queries
	 */
	public String generateQuery(int nrSimilar) {
		String result = "";
		boolean isWildcard = false;
		int nrTables = 0;
		if (Math.random() < oneTable) {
			nrTables = 1;
		} else {
			nrTables = 2;
		}
		ArrayList<Table> fromTables = new ArrayList<Table>();
		ArrayList<String> fromTablesString = new ArrayList<String>();
		ArrayList<String> selectItems = new ArrayList<String>();
		ArrayList<String> orderByItems = new ArrayList<String>();
		ArrayList<String> groupByItems = new ArrayList<String>();
		String[] whereClause = new String[nrSimilar];
		for (int i = 0; i < whereClause.length; i++) {
			whereClause[i] = "";
		}
		// generating tables
		for (int i = 0; i < nrTables; i++) {
			int j = (int)(Math.random()*tables.size());
			if (!fromTablesString.contains(tables.get(j).getTableName())) {
				fromTables.add(tables.get(j));
				fromTablesString.add(tables.get(j).getTableName());
			} else {
				i--;
			}
		}
		// generating selectitems
		boolean hasAggregate = false;
		if (Math.random() < wildCard) {
			isWildcard = true;
		} else {
			int nrSelect = (int)(Math.random()*maxSelect)+1;
			ArrayList<Column> possibleSelectItems = new ArrayList<Column>();
			for (int i = 0; i < fromTables.size(); i++) {
				possibleSelectItems.addAll(fromTables.get(i).getColumns());
			}
			if (nrSelect > possibleSelectItems.size()) {
				nrSelect = possibleSelectItems.size();
			}
			for(int i = 0; i < nrSelect; i++) {
				int j = (int)(Math.random()*possibleSelectItems.size());
				String selectItemString = "";
				if (Math.random() < aggragateProb && !hasAggregate) {
					int k = (int)(Math.random()*aggregate.length);
					hasAggregate = true;
					selectItemString = aggregate[k] + "(" + possibleSelectItems.get(j).getColumnName() + ")";
				} else {
					selectItemString = possibleSelectItems.get(j).getColumnName();
				}
				if (!selectItems.contains(selectItemString)) {
					selectItems.add(selectItemString);
				} else {
					i--;
				}
			}
		}
		// generating whereclause
		boolean isWhere = Math.random() < where;
		if (isWhere) {
			int nrWhere = (int)(Math.random()*maxWhere)+1;
			if (nrWhere < minWhere) {
				nrWhere = minWhere;
			}
			ArrayList<Column> possibleWhereElements = new ArrayList<Column>();
			for (int i = 0; i < fromTables.size(); i++) {
				possibleWhereElements.addAll(fromTables.get(i).getColumns());
			}
			if (nrWhere > possibleWhereElements.size()) {
				nrWhere = possibleWhereElements.size();
			}
			String joinString = "";
			boolean joinGenerated = false;
			// join
			if ((Math.random() < join) && (fromTables.size() == 2)) {
				ArrayList<Column> columnsTab1 = fromTables.get(0).getColumns();
				ArrayList<Column> columnsTab2 = fromTables.get(1).getColumns();
				for (int i = 0; i < columnsTab1.size(); i++) {
					int pos1 = columnsTab1.get(i).getColumnName().indexOf("_");
					String col1 = columnsTab1.get(i).getColumnName().substring(pos1+1);
					for (int j = 0; j < columnsTab2.size(); j++) {
						int pos2 = columnsTab2.get(j).getColumnName().indexOf("_");
						String col2 = columnsTab2.get(j).getColumnName().substring(pos2+1);
						if (col1.equals(col2)) {
							joinGenerated = true;
							joinString = " " + columnsTab1.get(i).getColumnName() + "=" + columnsTab2.get(j).getColumnName();
							if (nrWhere > 0) {
								joinString += " OR";
							}
						}
					}
				}
			}
			for (int i = 0; i < nrWhere; i++) {
				int j = (int)(Math.random()*possibleWhereElements.size());
				Column whereElement = possibleWhereElements.get(j);
				int operator = (int)(Math.random()*operators.length);
				int logic = (int)(Math.random()*logical.length);
				boolean isNot = (Math.random() < not);
				for (int k = 0; k < nrSimilar; k ++) {
					String operatorString = operators[operator];
					int valueIndex = (int)(Math.random()*whereElement.getValues().size());
					String valueString = whereElement.getValues().get(valueIndex);
					String whereElementString = whereElement.getColumnName() + operatorString + valueString;
					if (isNot) {
						whereElementString = "NOT(" + whereElementString + ")";
					}
					
					if (i < nrWhere-1) {
						whereClause[k] += " " + whereElementString + " " + logical[logic];
					} else {
						whereClause[k] += " " + whereElementString;
					}
					if (joinGenerated && i == 0) {
						whereClause[k] = joinString + whereClause[k];
					}
				}
			}
		}
		// generating groupby
		if (hasAggregate) {
			for (int i = 0; i < selectItems.size(); i++) {
				if (!selectItems.get(i).contains("(")) {
					groupByItems.add(selectItems.get(i));
				}
			}
		} else if (!isWildcard && Math.random() < groupBy && !hasAggregate) {
			int nrGroupBy = (int)(Math.random()*selectItems.size())+1;
			for (int i = 0; i < nrGroupBy; i++) {
				int j = (int)(Math.random()*selectItems.size());
				if (!groupByItems.contains(selectItems.get(j))) {
					groupByItems.add(selectItems.get(j));
				} else {
					i--;
				}
			}
		}		
		// generating orderby
		if (!isWildcard && Math.random() < orderBy) {
			int nrOrderBy = (int)(Math.random()*selectItems.size())+1;
			for (int i = 0; i < nrOrderBy; i++) {
				int j = (int)(Math.random()*selectItems.size());
				if (!orderByItems.contains(selectItems.get(j))) {
					orderByItems.add(selectItems.get(j));
				} else {
					i--;
				}
			}
		}
		// generating querystring
		for (int k = 0; k < nrSimilar; k++) {
			result += "SELECT ";
			if (isWildcard) {
				result += "*";
			} else {
				for (int i = 0; i < selectItems.size(); i++) {
					result += selectItems.get(i) + ", ";
				}
				result = result.substring(0, result.length()-2);
			}
			result += " FROM ";
			for (int i = 0; i < fromTables.size(); i++) {
				result += fromTables.get(i).getTableName() + ", ";
			}
			result = result.substring(0, result.length()-2);
			if (whereClause[k] != null && whereClause[k].length() != 0) {
				result += " WHERE" + whereClause[k];
			}
			if (groupByItems.size() > 0) {
				result += " GROUP BY ";
				for (int i = 0; i < groupByItems.size(); i++) {
					result += groupByItems.get(i) + ", ";
				}
				result = result.substring(0, result.length()-2);
			}
			if (orderByItems.size() > 0) {
				result += " ORDER BY ";
				for (int i = 0; i < orderByItems.size(); i++) {
					result += orderByItems.get(i) + ", ";
				}
				result = result.substring(0, result.length()-2);
			}
			result += ";\n";
			generatedQueries += 1;
			if (!isWhere || generatedQueries == nrQueries) {
				break;
			}
		}
		result = result.substring(0, result.length()-1);
		return result;
		
	}
	
	/**
	 * Generating output-file.
	 * 
	 */
	public void generateOutput() {
		try {
			File outputFile = new File(outputFileName);
			FileWriter out = new FileWriter(outputFile);
			while (generatedQueries < nrQueries) {
				out.write(generateQuery(nrSimilarQueries)+"\n");
			}
			out.write("QUIT;\n");
			out.close();
		} catch (IOException e) {
			System.out.println("IOEXception " + e.getMessage());
		}
	}
	/**
	 * Main-Method generates output-file.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		QueryGen qGen = new QueryGen();
		qGen.readTableInformation();
		System.out.println(qGen.tables);
		qGen.generateOutput();
	}

}
