package sqlindexwizard.workload;

import java.util.ArrayList;
import java.util.Vector;

import org.apache.log4j.Logger;

import sqlindexwizard.application.Input;
import sqlindexwizard.filter.EmptyWhereElement;
import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.filter.JoinElement;
import sqlindexwizard.filter.SelectElement;
import sqlindexwizard.filter.WhereElement;

import Zql.*;

/**
 * Represents an SelectQuery.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class SelectQuery extends Query {

	private ArrayList<SelectElement> select = new ArrayList<SelectElement>();
	private ArrayList<String> tables = new ArrayList<String>();
	private WhereElement where = null;
	private ArrayList<ItemWhereElement> itemList = new ArrayList<ItemWhereElement>();
	private ArrayList<JoinElement> joinList = new ArrayList<JoinElement>();
	private ArrayList<SelectElement> orderBy = new ArrayList<SelectElement>();
	private ArrayList<SelectElement> groupBy = new ArrayList<SelectElement>();
	private String having = "";
	private boolean isValid = true;
	
	private static Logger logger = Logger.getLogger(SelectQuery.class);
	
	/**
	 * Constructor, generates select query from statement.
	 * 
	 * @param st Statement
	 */
	public SelectQuery(ZStatement st) {
		super();
		extractTables((ZQuery)st);
		extractSelect((ZQuery)st);
		extractWhere((ZQuery)st);
		where = where.removePar();
		if (Input.toCNF) {
			where = where.toCNF();
		}
		if (Input.toDNF) {
			where = where.toDNF();
		}
		if (Input.simplify) {
			where = where.simplify();
		}
		extractOrderBy((ZQuery)st);
		extractGroupBy((ZQuery)st);
		generateItemList();
		generateJoinList();
		generateOtherList();
		logger.debug("Query ItemList: " + this.getItemList());
		logger.debug("Query JoinList: " + this.getJoinList());
		logger.debug("Generated new SelectQuery: " + this.toString());
	}
	
	/**
	 * Constructor calls super constructor.
	 * 
	 */
	public SelectQuery() {
		super();
	}
	
	/**
	 * Extracts select items from the query.
	 *  
	 * @param select query
	 */
	@SuppressWarnings("unchecked")
	public void extractSelect(ZQuery select) {
		Vector<ZSelectItem> selectvector = select.getSelect();
		for (int i = 0; i < selectvector.size(); i++) {
			ZSelectItem selectItem = selectvector.get(i);
			ZExp selectExp = selectItem.getExpression();
			if (selectExp instanceof ZQuery) {
				logger.info("No SubQuery in select clause supported!");
				isValid = false;
			}
			SelectElement selectElem = new SelectElement(selectItem.toString());
			selectElem.setPossibleTables(tables);
			this.select.add(selectElem);
		}
	}

	/**
	 * Extracts tables from the from-part.
	 * 
	 * @param select query
	 */
	@SuppressWarnings("unchecked")
	public void extractTables(ZQuery select) {
		Vector<ZFromItem> tablevector = select.getFrom();
		for (int i = 0; i < tablevector.size(); i++) {
			tables.add(tablevector.get(i).getTable().toUpperCase());
		}
	}
	
	/**
	 * Extracts where part.
	 * 
	 * @param select query
	 */
	public void extractWhere(ZQuery select) {
		ZExp whereclause = select.getWhere();
		if (whereclause instanceof ZQuery) {
			logger.info("No Subquerys supported!");
			isValid = false;
		} else if (whereclause instanceof ZExpression) {
			where = WhereElement.parseWhereClause(whereclause);	
			if (where == null) {
				logger.info("No Subquerys in where-clause supported!");
				isValid = false;
			}
		} else {
			where = new EmptyWhereElement();
		}
	}
	
	/**
	 * Extracts order by part.
	 * 
	 * @param select query
	 */
	@SuppressWarnings("unchecked")
	public void extractOrderBy(ZQuery select) {
		Vector<ZOrderBy> orderBy = select.getOrderBy();
		if (orderBy == null) {
			return;
		}
		for (int i = 0; i < orderBy.size(); i++) {
			ZExp orderByItem = orderBy.get(i).getExpression();
			if (orderByItem instanceof ZConstant) {
				SelectElement selectElem = new SelectElement(((ZConstant) orderByItem).getValue().toUpperCase());
				selectElem.setPossibleTables(tables);
				this.orderBy.add(selectElem);
			} else if (orderByItem instanceof ZExpression) {
				SelectElement selectElem = new SelectElement(((ZExpression)orderByItem).getOperand(0).toString());
				selectElem.setAggregate(((ZExpression)orderByItem).getOperator().toUpperCase());
				selectElem.setPossibleTables(tables);
				this.orderBy.add(selectElem);
			} else if (orderByItem instanceof ZQuery) {
				logger.info("No Subquerys supported!");
				isValid = false;
			}	
		}
	}
	
	/**
	 * Extracts group by part.
	 * 
	 * @param select query
	 */
	@SuppressWarnings("unchecked")
	public void extractGroupBy(ZQuery select) {
		ZGroupBy groupBy = select.getGroupBy();
		if (groupBy == null) {
			return;
		}
		Vector<ZExp> groupByElements = groupBy.getGroupBy();
		if (groupBy.getHaving() != null) {
			this.having = groupBy.getHaving().toString();
		}
		for (int i = 0; i < groupByElements.size(); i++) {
			ZExp groupByItem = groupByElements.get(i);
			if (groupByItem instanceof ZConstant) {
				SelectElement selectElem = new SelectElement(((ZConstant) groupByItem).getValue().toUpperCase());
				selectElem.setPossibleTables(tables);
				this.groupBy.add(selectElem);
			} else if (groupByItem instanceof ZExpression) {
				SelectElement selectElem = new SelectElement(((ZExpression)groupByItem).getOperand(0).toString());
				selectElem.setAggregate(((ZExpression)groupByItem).getOperator().toUpperCase());
				selectElem.setPossibleTables(tables);
				this.groupBy.add(selectElem);
			} else if (groupByItem instanceof ZQuery) {
				logger.info("No Subquerys supported!");
				isValid = false;
			}
		}
	}
	
	/**
	 * Converts where part into DNF.
	 * 
	 */
	public void whereClauseToDNF(){
		where = where.toDNF();
	}
	
	/**
	 * Converts where part into CNF.
	 */
	public void whereClauseToCNF(){
		where = where.toCNF();
	}
	
	/**
	 * Simplyfies where part.
	 * 
	 */
	public void whereClauseSimplify(){
		where = where.simplify();
	}
	
	/**
	 * Generates the itemList from the where-part, tries to eliminate
	 * duplicates.
	 * Also adds them to the Workload itemList.
	 * 
	 */
	public void generateItemList() {
		ArrayList<ItemWhereElement> temp = where.getItemList();
		ArrayList<ItemWhereElement> result = new ArrayList<ItemWhereElement>();
		for(int i = 0; i < temp.size(); i++) {
			if (temp.get(i).isJoin() == false) {
				if (temp.get(i).hasTable()) {
					result.add(temp.get(i));
					Workload.getItemList().add(temp.get(i));
				} else {
					temp.get(i).setPossibleTables(tables);
					result.add(temp.get(i));
					Workload.getItemList().add(temp.get(i));
				}
			}
		}
		itemList = result;
	}
	
	/**
	 * Extracting joins from the where-clause and adds them to the
	 * joinList.
	 * Also adds them to the Workload joinList.
	 * 
	 */
	public void generateJoinList() {
		ArrayList<ItemWhereElement> temp = where.getItemList();
		ArrayList<JoinElement> result = new ArrayList<JoinElement>();
		for(int i = 0; i < temp.size(); i++) {
			if (temp.get(i).isJoin() == true) {
				JoinElement elem = temp.get(i).getJoinElement();
				if (elem.hasTableLeft() && elem.hasTableRight()) {
					result.add(elem);
					Workload.getJoinList().add(elem);
				} else {
					temp.get(i).setPossibleTables(tables);
					elem.setPossibleTables(tables);
					result.add(elem);
					Workload.getJoinList().add(elem);
				}
			}
		}
		joinList = result;
	}
	
	/**
	 * Adding SelectElements to the Workloads static otherList.
	 * Will be needed for generating items in advise_index.
	 * 
	 */
	public void generateOtherList() {
		Workload.getOtherList().addAll(orderBy);
		Workload.getOtherList().addAll(select);
		Workload.getOtherList().addAll(groupBy);
	}

	/**
	 * Getting itemList.
	 * 
	 * @return the itemList
	 */
	public ArrayList<ItemWhereElement> getItemList() {
		return itemList;
	}

	/**
	 * Getting joinList.
	 * 
	 * @return the joinList
	 */
	public ArrayList<JoinElement> getJoinList() {
		return joinList;
	}

	/**
	 * Getting tables.
	 * 
	 * @return tables.
	 */
	public ArrayList<String> getTables() {
		return tables;
	}
	
	/**
	 * Getting select.
	 * 
	 * @return the select
	 */
	public ArrayList<SelectElement> getSelect() {
		return select;
	}

	/**
	 * Getting the where-clause.
	 * 
	 * @return where-clause
	 */
	public WhereElement getWhere() {
		return where;
	}

	/**
	 * Getting orderBy.
	 * 
	 * @return the orderBy
	 */
	public ArrayList<SelectElement> getOrderBy() {
		return orderBy;
	}

	/**
	 * Getting groupBy.
	 * 
	 * @return the groupBy
	 */
	public ArrayList<SelectElement> getGroupBy() {
		return groupBy;
	}

	/**
	 * Getting isValid.
	 * 
	 * @return the isValid
	 */
	public boolean isValid() {
		return isValid;
	}

	/**
	 * Getting the having-part.
	 * 
	 * @return having-part
	 */
	public String getHaving() {
		return having;
	}

	/**
	 * Clones this SelectQuery.
	 * 
	 * @return clone
	 */
	@SuppressWarnings("unchecked")
	public SelectQuery clone() {
		SelectQuery clone = new SelectQuery();
		clone.tables = (ArrayList<String>) this.tables.clone();
		clone.where = this.where.clone();
		clone.select = (ArrayList<SelectElement>) this.select.clone();
		clone.orderBy = (ArrayList<SelectElement>) this.orderBy.clone();
		clone.groupBy = (ArrayList<SelectElement>) this.groupBy.clone();
		clone.itemList = (ArrayList<ItemWhereElement>) this.itemList.clone();
		clone.joinList = (ArrayList<JoinElement>) this.joinList.clone();
		clone.having = this.having;
		clone.isValid = this.isValid;
		return clone;
	}

	/**
	 * String-Representation of Select query.
	 * 
	 * @return String-Representation
	 */
	public String toString() {
		String result = "SELECT ";
		for(SelectElement selectItem: select) {
			result += selectItem + ", ";
		}
		result = result.substring(0, result.length()-2);
		result += " FROM ";
		for(String table: tables) {
			result += table + ", ";
		}
		result = result.substring(0, result.length()-2);
		if (!(where instanceof EmptyWhereElement )) {
			result += " WHERE " + where.toString();
		}
		if (groupBy.size() != 0) {
			result += " GROUP BY ";
			for(SelectElement gB: groupBy) {
				result += gB + ", ";
			}
			result = result.substring(0, result.length()-2);
		}
		if (!having.equals("")) {
			result += " HAVING " + having;
		}
		if (orderBy.size() != 0) {
			result += " ORDER BY ";
			for(SelectElement oB: orderBy) {
				result += oB + ", ";
			}
			result = result.substring(0, result.length()-2);
		}
		result += ";";
		return result;	
	}
	
}
