package sqlindexwizard.workload;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.log4j.Logger;

import sqlindexwizard.clustering.TableElement;
import sqlindexwizard.filter.ItemWhereElement;
import sqlindexwizard.filter.JoinElement;
import sqlindexwizard.filter.SelectElement;

import Zql.*;
/**
 * Represents a workload. Reads Queries from StdIn or from files an generates Query-Lists.
 * 
 * @author Benjamin Daeumlich
 *
 */
public class Workload {
	
	private ListOfQueries selectQueryList = new ListOfQueries();
	private ListOfQueries updateQueryList = new ListOfQueries();
	private ListOfQueries insertQueryList = new ListOfQueries();
	private ListOfQueries deleteQueryList = new ListOfQueries();
	private ListOfQueries notAddedQueryList = new ListOfQueries();
	
	private static HashMap<String,TableElement> mapOfTableElements = new HashMap<String,TableElement>();
	private static ArrayList<ItemWhereElement> itemList = new ArrayList<ItemWhereElement>();
	private static ArrayList<JoinElement> joinList = new ArrayList<JoinElement>();
	private static ArrayList<SelectElement> otherList = new ArrayList<SelectElement>();
	private static HashMap<ItemWhereElement, Double> mapOfSelectivities = new HashMap<ItemWhereElement, Double>();
	
	private int nrOfNotAddedQueries = 0;

	private static Logger logger = Logger.getLogger(Workload.class);
	
	/**
	 * Constructor, standard.
	 */
	public Workload() {
		//nothing to do
	}
	
	/**
	 * Constructor, generates Workload from StdIn.
	 */
	@SuppressWarnings("deprecation")
	public Workload(int i) {
		try {
			logger.info("Generating Workload from StdIn.");
			DataInputStream s = new DataInputStream(System.in);
			byte[] stringBytes = s.readLine().getBytes();
			while (stringBytes != null) {
				ByteArrayInputStream input = new ByteArrayInputStream(stringBytes);
				ZqlParser p = new ZqlParser(input);
				ZStatement st = null;
				try {
					st = p.readStatement();
					if (st == null) {
						break;
					}
					Query q = Query.generateQuery(st);
					if (this.addQuery(q)) {
						logger.info("Added Query " + q + " successful!");
					} else {
						logger.info("Failed to add Query: " + new String(stringBytes));
						notAddedQueryList.add(new InvalidQuery(new String(stringBytes)));
						nrOfNotAddedQueries++;
					}
				} catch (ParseException e){
					notAddedQueryList.add(new InvalidQuery(new String(stringBytes)));
					nrOfNotAddedQueries++;
					logger.error("Trying to Parse invalid SQL-Statement: " + new String(stringBytes));
				}
				stringBytes = s.readLine().getBytes();
			}
		} catch (IOException e){
			logger.error("IOException: " + e);
		}
	}
	
	/**
	 * Constructor, generates Workload from file s.
	 * 
	 * @param s LogFile
	 */
	public Workload(String s) {
		logger.info("Adding logfile to workload: " + s);
		File f = new File(s);
		generateWorkloadFromFile(f);
	}
	
	/**
	 * Constructor, generates Workload from files in s.
	 * 
	 * @param s LogFile array
	 */
	public Workload(String[] s) {
		for (int i = 0; i < s.length; i++) {
			logger.info("Adding logfile to workload: " + s[i]);
			File f = new File(s[i]);
			generateWorkloadFromFile(f);
		}
	}
	
	/**
	 * Reads statements from file, generates and adds queries to the QueryLists.
	 * 
	 * @param f File
	 */
	@SuppressWarnings("deprecation")
	public void generateWorkloadFromFile(File f) {
		try {
			logger.info("Generating Workload from file " + f.getName() + ".");
			DataInputStream s = new DataInputStream(new FileInputStream(f));
			byte[] stringBytes = s.readLine().getBytes();
			while (stringBytes != null) {
				ByteArrayInputStream input = new ByteArrayInputStream(stringBytes);
				ZqlParser p = new ZqlParser(input);
				p.addCustomFunction("FIRST", 1);
				p.addCustomFunction("LAST", 1);
				ZStatement st = null;
				try {
					st = p.readStatement();
					if (st == null) {
						break;
					}
					Query q = Query.generateQuery(st);
					if (this.addQuery(q)) {
						logger.debug("Added Query " + q + " successful!");
					} else {
						logger.debug("Failed to add Query: " + new String(stringBytes));
						notAddedQueryList.add(new InvalidQuery(new String(stringBytes)));
						nrOfNotAddedQueries++;
					}
				} catch (ParseException e){
					notAddedQueryList.add(new InvalidQuery(new String(stringBytes)));
					nrOfNotAddedQueries++;
					logger.error("Trying to Parse invalid SQL-Statement: " + new String(stringBytes));
				}
				stringBytes = s.readLine().getBytes();
			}
			logger.info("Finished generating Workload from file " + f.getName() + ".");
		} catch (FileNotFoundException e) {
			logger.error("File Not Found: " + f.toString());
		} catch (IOException e){
			logger.error("IOException: " + e);
		}
	}
	
	/**
	 * Adds query to the correct QueryList.
	 * 
	 * @param Q Query to add
	 * @return true if added successfully
	 */
	public boolean addQuery(Query Q) {
		boolean result;
		if (Q instanceof SelectQuery) {
			result = selectQueryList.add(Q);
		} else if (Q instanceof InsertQuery) {
			result = insertQueryList.add(Q);
		} else if (Q instanceof DeleteQuery) {
			result = deleteQueryList.add(Q);
		} else if (Q instanceof UpdateQuery) {
			result = updateQueryList.add(Q);
		} else if (Q instanceof InvalidQuery) {
			result = notAddedQueryList.add(Q);
			nrOfNotAddedQueries++;
		} else {
			result = false;
		}
		return result;
	}
	
	/**
	 * Adds a ListOfQuerys to the correct querylists.
	 * 
	 * @param QL ListOfQueries to add
	 * @return true if all queries added successfully
	 */
	public boolean addAllQueries(ListOfQueries QL) {
		boolean result = false;
		if (QL == null) {
			return false;
		}
		Iterator<Query> it = QL.iterator();
		while (it.hasNext()) {
			result = result & addQuery(it.next());
		}
		return result;
	}
	
	/**
	 * Gets nrOfTotalAddedQueries.
	 * 
	 * @return nrOfAddedQueries
	 */
	public int getNrOfAddedQueries() {
		return selectQueryList.size() + insertQueryList.size() + updateQueryList.size() + deleteQueryList.size();
	}
	
	/**
	 * Getting nrOfNotAddedQueries.
	 * 
	 * @return nrOfNotAddedQueries
	 */
	public int getNrOfNotAddedQueries() {
		return nrOfNotAddedQueries;
	}
	
	/**
	 * Getting totalNorOfQueries.
	 * 
	 * @return totalNrOfQueries
	 */
	public int getTotalNrOfQueries() {
		return getNrOfAddedQueries() + getNrOfNotAddedQueries();
	}

	/**
	 * Getting List of SelectQueries.
	 * 
	 * @return ListOfSelectQueries.
	 */
	public ListOfQueries getSelectQueryList() {
		return selectQueryList;
	}

	/**
	 * Setting List of SelectQueries.
	 * 
	 * @param selectQueryList list to set
	 */
	public void setSelectQueryList(ListOfQueries selectQueryList) {
		this.selectQueryList = selectQueryList;
	}
	
	/**
	 * Getting List of UpdateQueries.
	 * 
	 * @return the updateQueryList
	 */
	public ListOfQueries getUpdateQueryList() {
		return updateQueryList;
	}

	/**
	 * Setting List of UpdateQueries.
	 * 
	 * @param updateQueryList the updateQueryList to set
	 */
	public void setUpdateQueryList(ListOfQueries updateQueryList) {
		this.updateQueryList = updateQueryList;
	}

	/**
	 * Getting List of InsertQueries.
	 * 
	 * @return the insertQueryList
	 */
	public ListOfQueries getInsertQueryList() {
		return insertQueryList;
	}

	/**
	 * Setting List of InsertQueries.
	 * 
	 * @param insertQueryList the insertQueryList to set
	 */
	public void setInsertQueryList(ListOfQueries insertQueryList) {
		this.insertQueryList = insertQueryList;
	}

	/**
	 * Getting List of DeleteQueries.
	 * 
	 * @return the deleteQueryList
	 */
	public ListOfQueries getDeleteQueryList() {
		return deleteQueryList;
	}

	/**
	 * Setting List of DelectQueries.
	 * 
	 * @param deleteQueryList the deleteQueryList to set
	 */
	public void setDeleteQueryList(ListOfQueries deleteQueryList) {
		this.deleteQueryList = deleteQueryList;
	}

	/**
	 * Getting the notAddedQueryList.
	 * 
	 * @return the notAddedQueryList
	 */
	public ListOfQueries getNotAddedQueryList() {
		return notAddedQueryList;
	}

	/**
	 * Setting the notAddedQueryList.
	 * 
	 * @param notAddedQueryList the notAddedQueryList to set
	 */
	public void setNotAddedQueryList(ListOfQueries notAddedQueryList) {
		this.notAddedQueryList = notAddedQueryList;
	}

	/**
	 * Getting Map of TableElements.
	 * 
	 * @return MapOfTableElements
	 */
	public static HashMap<String, TableElement> getMapOfTableElements() {
		return mapOfTableElements;
	}

	/**
	 * Setting Map of TableElements.
	 * 
	 * @param mapOfTableElements new Map
	 */
	public static void setMapOfTableElements(
			HashMap<String, TableElement> mapOfTableElements) {
		Workload.mapOfTableElements = mapOfTableElements;
	}

	/**
	 * Getting ItemList.
	 * 
	 * @return ItemList
	 */
	public static ArrayList<ItemWhereElement> getItemList() {
		return itemList;
	}

	/**
	 * Setting ItemList.
	 * 
	 * @param itemList new ItemList
	 */
	public static void setItemList(ArrayList<ItemWhereElement> itemList) {
		Workload.itemList = itemList;
	}

	/**
	 * Getting the joinList.
	 * 
	 * @return the joinList
	 */
	public static ArrayList<JoinElement> getJoinList() {
		return joinList;
	}

	/**
	 * Setting the joinList.
	 * 
	 * @param joinList the joinList to set
	 */
	public static void setJoinList(ArrayList<JoinElement> joinList) {
		Workload.joinList = joinList;
	}

	/**
	 * Getting the otherList.
	 * 
	 * @return the otherList
	 */
	public static ArrayList<SelectElement> getOtherList() {
		return otherList;
	}

	/**
	 * Setting the otherList.
	 * 
	 * @param otherList the otherList to set
	 */
	public static void setOtherList(ArrayList<SelectElement> otherList) {
		Workload.otherList = otherList;
	}

	/**
	 * Getting Map of Selectivities.
	 * 
	 * @return Map of Selectivities
	 */
	public static HashMap<ItemWhereElement, Double> getMapOfSelectivities() {
		return mapOfSelectivities;
	}

	/**
	 * Setting Map of Selectivities.
	 * 
	 * @param mapOfSelectivities new Map
	 */
	public static void setMapOfSelectivities(
			HashMap<ItemWhereElement, Double> mapOfSelectivities) {
		Workload.mapOfSelectivities = mapOfSelectivities;
	}
	
	/**
	 * Clones this Workload.
	 */
	public Workload clone() {
		Workload clone = new Workload();
		clone.selectQueryList = (ListOfQueries)this.selectQueryList.clone();
		clone.updateQueryList = (ListOfQueries)this.updateQueryList.clone();
		clone.deleteQueryList = (ListOfQueries)this.deleteQueryList.clone();
		clone.insertQueryList = (ListOfQueries)this.insertQueryList.clone();
		clone.notAddedQueryList = (ListOfQueries)this.notAddedQueryList.clone();
		clone.nrOfNotAddedQueries = this.nrOfNotAddedQueries;
		return clone;
	}
	
	/**
	 * Detailed String Representation of the Workload.
	 * 
	 * @return string representation
	 */
	public String toStringDetail() {
		StringBuffer result = new StringBuffer();
		result.append("selectQueryList:\n" + selectQueryList + "\n");
		result.append("updateQueryList:\n" + updateQueryList + "\n");
		result.append("deleteQueryList:\n" + deleteQueryList + "\n");
		result.append("insertQueryList:\n" + insertQueryList + "\n");
		result.append("notAddedQueryList:\n" + notAddedQueryList + "\n");
		result.append("nrAddedQueries: " + getNrOfAddedQueries() + "\n");
		result.append("nrNotAddedQueries: " + getNrOfNotAddedQueries() + "\n");
		result.append("nrQueries: " + getTotalNrOfQueries() + "\n");
		return result.toString();
	}
	
	/**
	 * String Representation of the Workload.
	 * 
	 * @return string representation
	 */
	public String toString() {
		StringBuffer result = new StringBuffer();
		result.append(selectQueryList);
		result.append(updateQueryList);
		result.append(deleteQueryList);
		result.append(insertQueryList);
		result.append(notAddedQueryList);
		result.append("\n\tnrQueries: " + getTotalNrOfQueries() + "\n");
		return result.toString();
	}
	
}
